"""
This is a non-official implementation of 'Client Selection in Federated Learning:
Convergence Analysis and Power-of-Choice Selection Strategies' (https://arxiv.org/abs/2010.01243).
"""
import numpy as np
from flgo.algorithm.fedavg import Client
from flgo.algorithm.fedbase import BasicServer
import flgo.simulator.base as ss

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'d': self.num_clients})

    @ss.with_availability
    def sample(self):
        # create candidate set A
        num_candidate = min(self.d, len(self.available_clients))
        p_candidate = np.array([len(self.clients[cid].train_data) for cid in self.available_clients])
        candidate_set = np.random.choice(self.available_clients, num_candidate, p=p_candidate / p_candidate.sum(), replace=False)
        candidate_set = sorted(candidate_set)
        # communicate with the candidates for their local_movielens_recommendation loss
        losses = []
        for cid in candidate_set:
            losses.append(self.clients[cid].test(self.model)['loss'])
        # sort candidate set according to their local_movielens_recommendation loss value, and choose the top-M highest ones
        sort_id = np.array(losses).argsort().tolist()
        sort_id.reverse()
        num_selected = min(self.clients_per_round, len(self.available_clients))
        selected_clients = np.array(self.available_clients)[sort_id][:num_selected]
        return selected_clients.tolist()


import datetime
import errno
import os
import time
from collections import defaultdict, deque
import flgo.benchmark.toolkits.cv.detection.presets as presets
import torch
import torchvision
import torch.distributed as dist

def get_transform(train, data_augmentation:str='hflip'):
    if train:
        return presets.DetectionPresetTrain(data_augmentation=data_augmentation)
    else:
        return presets.DetectionPresetEval()

class SmoothedValue:
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """

    def __init__(self, window_size=20, fmt=None):
        if fmt is None:
            fmt = "{median:.4f} ({global_avg:.4f})"
        self.deque = deque(maxlen=window_size)
        self.total = 0.0
        self.count = 0
        self.fmt = fmt

    def update(self, value, n=1):
        self.deque.append(value)
        self.count += n
        self.total += value * n

    def synchronize_between_processes(self):
        """
        Warning: does not synchronize the deque!
        """
        if not is_dist_avail_and_initialized():
            return
        t = torch.tensor([self.count, self.total], dtype=torch.float64, device="cuda")
        dist.barrier()
        dist.all_reduce(t)
        t = t.tolist()
        self.count = int(t[0])
        self.total = t[1]

    @property
    def median(self):
        d = torch.tensor(list(self.deque))
        return d.median().item()

    @property
    def avg(self):
        d = torch.tensor(list(self.deque), dtype=torch.float32)
        return d.mean().item()

    @property
    def global_avg(self):
        return self.total / self.count

    @property
    def max(self):
        return max(self.deque)

    @property
    def value(self):
        return self.deque[-1]

    def __str__(self):
        return self.fmt.format(
            median=self.median, avg=self.avg, global_avg=self.global_avg, max=self.max, value=self.value
        )


def all_gather(data):
    """
    Run all_gather on arbitrary picklable data (not necessarily tensors)
    Args:
        data: any picklable object
    Returns:
        list[data]: list of data gathered from each rank
    """
    world_size = get_world_size()
    if world_size == 1:
        return [data]
    data_list = [None] * world_size
    dist.all_gather_object(data_list, data)
    return data_list


def reduce_dict(input_dict, average=True):
    """
    Args:
        input_dict (dict): all the values will be reduced
        average (bool): whether to do average or sum
    Reduce the values in the dictionary from all processes so that all processes
    have the averaged results. Returns a dict with the same fields as
    input_dict, after reduction.
    """
    world_size = get_world_size()
    if world_size < 2:
        return input_dict
    with torch.inference_mode():
        names = []
        values = []
        # sort the keys so that they are consistent across processes
        for k in sorted(input_dict.keys()):
            names.append(k)
            values.append(input_dict[k])
        values = torch.stack(values, dim=0)
        dist.all_reduce(values)
        if average:
            values /= world_size
        reduced_dict = {k: v for k, v in zip(names, values)}
    return reduced_dict


class MetricLogger:
    def __init__(self, delimiter="\t"):
        self.meters = defaultdict(SmoothedValue)
        self.delimiter = delimiter

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, torch.Tensor):
                v = v.item()
            assert isinstance(v, (float, int))
            self.meters[k].update(v)

    def __getattr__(self, attr):
        if attr in self.meters:
            return self.meters[attr]
        if attr in self.__dict__:
            return self.__dict__[attr]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

    def __str__(self):
        loss_str = []
        for name, meter in self.meters.items():
            loss_str.append(f"{name}: {str(meter)}")
        return self.delimiter.join(loss_str)

    def synchronize_between_processes(self):
        for meter in self.meters.values():
            meter.synchronize_between_processes()

    def add_meter(self, name, meter):
        self.meters[name] = meter

    def log_every(self, iterable, print_freq, header=None):
        i = 0
        if not header:
            header = ""
        start_time = time.time()
        end = time.time()
        iter_time = SmoothedValue(fmt="{avg:.4f}")
        data_time = SmoothedValue(fmt="{avg:.4f}")
        space_fmt = ":" + str(len(str(len(iterable)))) + "d"
        if torch.cuda.is_available():
            log_msg = self.delimiter.join(
                [
                    header,
                    "[{0" + space_fmt + "}/{1}]",
                    "eta: {eta}",
                    "{meters}",
                    "time: {time}",
                    "data: {data}",
                    "max mem: {memory:.0f}",
                ]
            )
        else:
            log_msg = self.delimiter.join(
                [header, "[{0" + space_fmt + "}/{1}]", "eta: {eta}", "{meters}", "time: {time}", "data: {data}"]
            )
        MB = 1024.0 * 1024.0
        for obj in iterable:
            data_time.update(time.time() - end)
            yield obj
            iter_time.update(time.time() - end)
            if i % print_freq == 0 or i == len(iterable) - 1:
                eta_seconds = iter_time.global_avg * (len(iterable) - i)
                eta_string = str(datetime.timedelta(seconds=int(eta_seconds)))
                if torch.cuda.is_available():
                    print(
                        log_msg.format(
                            i,
                            len(iterable),
                            eta=eta_string,
                            meters=str(self),
                            time=str(iter_time),
                            data=str(data_time),
                            memory=torch.cuda.max_memory_allocated() / MB,
                        )
                    )
                else:
                    print(
                        log_msg.format(
                            i, len(iterable), eta=eta_string, meters=str(self), time=str(iter_time), data=str(data_time)
                        )
                    )
            i += 1
            end = time.time()
        total_time = time.time() - start_time
        total_time_str = str(datetime.timedelta(seconds=int(total_time)))
        print(f"{header} Total time: {total_time_str} ({total_time / len(iterable):.4f} s / it)")


def collate_fn(batch):
    return tuple(zip(*batch))


def mkdir(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def setup_for_distributed(is_master):
    """
    This function disables printing when not in master process
    """
    import builtins as __builtin__

    builtin_print = __builtin__.print

    def print(*args, **kwargs):
        force = kwargs.pop("force", False)
        if is_master or force:
            builtin_print(*args, **kwargs)

    __builtin__.print = print


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()


def is_main_process():
    return get_rank() == 0


def save_on_master(*args, **kwargs):
    if is_main_process():
        torch.save(*args, **kwargs)


def init_distributed_mode(args):
    if "RANK" in os.environ and "WORLD_SIZE" in os.environ:
        args.rank = int(os.environ["RANK"])
        args.world_size = int(os.environ["WORLD_SIZE"])
        args.gpu = int(os.environ["LOCAL_RANK"])
    elif "SLURM_PROCID" in os.environ:
        args.rank = int(os.environ["SLURM_PROCID"])
        args.gpu = args.rank % torch.cuda.device_count()
    else:
        print("Not using distributed mode")
        args.distributed = False
        return

    args.distributed = True

    torch.cuda.set_device(args.gpu)
    args.dist_backend = "nccl"
    print(f"| distributed init (rank {args.rank}): {args.dist_url}", flush=True)
    torch.distributed.init_process_group(
        backend=args.dist_backend, init_method=args.dist_url, world_size=args.world_size, rank=args.rank
    )
    torch.distributed.barrier()
    setup_for_distributed(args.rank == 0)


from ..toolkits.nlp.classification import DataPipeTaskPipe, DataPipeGenerator, GeneralCalculator
import torchtext.datasets
import os
import flgo.benchmark

path = os.path.join(flgo.benchmark.path, 'RAW_DATA','AG_NEWS')

def build_datapipes(root:str, split:str='train'):
    dp = torchtext.datasets.AG_NEWS(root=root, split=split)
    return dp

class TaskGenerator(DataPipeGenerator):
    def __init__(self, rawdata_path:str=path):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1], rawdata_path=rawdata_path, build_datapipes=build_datapipes)
        self.build_datapipes = build_datapipes
        self.additional_option = {'root':self.rawdata_path}
        self.train_additional_option = {'split': 'train'}
        self.test_additional_option = {'split':'test'}

class TaskPipe(DataPipeTaskPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, build_datapipes)

TaskCalculator = GeneralCalculator


import json
import os
import flgo
import torch
import torch.nn as nn
import torch.nn.functional as F

from flgo.utils.fmodule import FModule


class Model(FModule):

    def __init__(self, hidden_size=50, num_features=8):
        super(Model, self).__init__()
        self.encoder = nn.GRU(input_size=num_features, hidden_size=hidden_size, num_layers=4)
        self.fc = nn.Linear(hidden_size * 4, num_features)

    def forward(self, x):  # [B, W, M]
        x = x.permute([1, 0, 2])  # [W, B, M]
        _, h = self.encoder(x)  # [4, B, h]
        h = h.permute([1, 0, 2]).contiguous()  # [B, 4, h]
        h = h.view(h.shape[0], -1)  # [B, 4*h]
        x = self.fc(h)  # [B, M]
        return F.sigmoid(x)


def init_local_module(object):
    pass


def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)



from __future__ import annotations

from flgo.benchmark.base import *
import torch
import os
from torch.utils.data.dataloader import DataLoader

class BuiltinClassGenerator(BasicTaskGenerator):
    r"""
    Generator for the time series dataset.

    Args:
        benchmark (str): the name of the benchmark
        rawdata_path (str): the path storing the raw data
        builtin_class (class): dataset class
    """
    def __init__(self, benchmark, rawdata_path, builtin_class):
        super().__init__(benchmark, rawdata_path)
        self.builtin_class = builtin_class
        self.additional_option = {}
        self.train_additional_option = {}
        self.test_additional_option = {}

    def load_data(self, *args, **kwargs):
        train_default_init_para = {'root': self.rawdata_path, 'train': True}
        test_default_init_para = {'root': self.rawdata_path, 'train': False}
        train_default_init_para.update(self.additional_option)
        train_default_init_para.update(self.train_additional_option)
        test_default_init_para.update(self.additional_option)
        test_default_init_para.update(self.test_additional_option)
        # if 'kwargs' not in self.builtin_class.__init__.__annotations__:
        #     train_pop_key = [k for k in train_default_init_para.keys() if
        #                      k not in self.builtin_class.__init__.__annotations__]
        #     test_pop_key = [k for k in test_default_init_para.keys() if
        #                     k not in self.builtin_class.__init__.__annotations__]
        #     for k in train_pop_key: train_default_init_para.pop(k)
        #     for k in test_pop_key: test_default_init_para.pop(k)
        # init datasets
        self.train_data = self.builtin_class(**train_default_init_para)
        self.test_data = self.builtin_class(**test_default_init_para)
        # handle train_data to triple (user, item, rating)
        # self.train_data = [tuple(self.train_data[idx][k] for k in self.train_data.enum_data) for idx in
        #                    range(len(self.train_data))]

    def partition(self):
        self.local_datas = self.partitioner(self.train_data)
        self.num_clients = len(self.local_datas)

class BuiltinClassPipe(BasicTaskPipe):
    r"""
    TaskPipe for the time series dataset.

    Args:
        task_path (str): the path of the task
        builtin_class (class): dataset class
    """

    class TaskDataset(torch.utils.data.Subset):
        def __init__(self, dataset, indices, perturbation=None, pin_memory=False):
            super().__init__(dataset, indices)
            self.dataset = dataset
            self.indices = indices
            self.perturbation = {idx: p for idx, p in zip(indices, perturbation)} if perturbation is not None else None
            self.pin_memory = pin_memory
            if not self.pin_memory:
                self.X = None
                self.Y = None
            else:
                self.X = torch.stack([self.dataset[i][0] for i in self.indices])
                self.Y = torch.Tensor([self.dataset[i][1].tolist() for i in self.indices])

        def __getitem__(self, idx):
            if self.X is not None:
                if self.perturbation is None:
                    return self.X[idx], self.Y[idx]
                else:
                    return self.X[idx] + self.perturbation[self.indices[idx]], self.Y[idx]
            else:
                if self.perturbation is None:
                    if isinstance(idx, list):
                        return self.dataset[[self.indices[i] for i in idx]]
                    return self.dataset[self.indices[idx]]
                else:
                    return self.dataset[self.indices[idx]][0] + self.perturbation[self.indices[idx]], \
                    self.dataset[self.indices[idx]][1]

    def __init__(self, task_path, buildin_class, transform=None):
        """
        Args:
            task_path (str): the path of the task
            builtin_class (class): class in torchvision.datasets
            transform (torchvision.transforms.*): the transform
        """
        super(BuiltinClassPipe, self).__init__(task_path)
        self.builtin_class = buildin_class
        self.transform = transform

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'server_data': list(range(len(generator.test_data))),  'rawdata_path': generator.rawdata_path, 'additional_option': generator.additional_option, 'train_additional_option':generator.train_additional_option, 'test_additional_option':generator.test_additional_option, }
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid],}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        train_default_init_para = {'root': self.feddata['rawdata_path'], 'train': True}
        test_default_init_para = {'root': self.feddata['rawdata_path'], 'train': False}
        if 'additional_option' in self.feddata.keys():
            train_default_init_para.update(self.feddata['additional_option'])
            test_default_init_para.update(self.feddata['additional_option'])
        if 'train_additional_option' in self.feddata.keys(): train_default_init_para.update(
            self.feddata['train_additional_option'])
        if 'test_additional_option' in self.feddata.keys(): test_default_init_para.update(
            self.feddata['test_additional_option'])
        # if 'kwargs' not in self.builtin_class.__init__.__annotations__:
        #     train_pop_key = [k for k in train_default_init_para.keys() if
        #                      k not in self.builtin_class.__init__.__annotations__]
        #     test_pop_key = [k for k in test_default_init_para.keys() if
        #                     k not in self.builtin_class.__init__.__annotations__]
        #     for k in train_pop_key: train_default_init_para.pop(k)
        #     for k in test_pop_key: test_default_init_para.pop(k)
        train_data = self.builtin_class(**train_default_init_para)
        test_data = self.builtin_class(**test_default_init_para)
        test_data = self.TaskDataset(test_data, list(range(len(test_data))), None, running_time_option['pin_memory'])
        # rearrange data for server
        server_data_test, server_data_valid = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'valid': server_data_valid}}
        # rearrange data for clients
        local_perturbation = self.feddata['local_perturbation'] if 'local_perturbation' in self.feddata.keys() else [None for _ in self.feddata['client_names']]
        for cid, cname in enumerate(self.feddata['client_names']):
            cpert = None if local_perturbation[cid] is None else [torch.tensor(t) for t in local_perturbation[cid]]
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'], cpert, running_time_option['pin_memory'])
            cdata_train, cdata_valid = self.split_dataset(cdata, running_time_option['train_holdout'])
            if running_time_option['train_holdout'] > 0 and running_time_option['local_test']:
                cdata_valid, cdata_test = self.split_dataset(cdata_valid, 0.5)
            else:
                cdata_test = None
            task_data[cname] = {'train': cdata_train, 'valid': cdata_valid, 'test': cdata_test}
        return task_data

class GeneralCalculator(BasicTaskCalculator):
    r"""
    Calculator for the time series dataset.

    Args:
        device (torch.device): device
        optimizer_name (str): the name of the optimizer
    """
    def __init__(self, device, optimizer_name='sgd'):
        super().__init__(device, optimizer_name)
        self.criterion = torch.nn.CrossEntropyLoss()
        self.DataLoader = DataLoader

    def compute_loss(self, model, data):
        tdata = self.to_device(data)
        outputs = model(tdata[0])
        loss = self.criterion(outputs, tdata[-1].long())
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        """
        Metric = [mean_accuracy, mean_loss]

        Args:
            model:
            dataset:
            batch_size:
        Returns: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size == -1: batch_size = len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)
        total_loss = 0.0
        num_correct = 0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data[0])
            batch_mean_loss = self.criterion(outputs, batch_data[-1].long()).item()
            y_pred = outputs.data.max(1, keepdim=True)[1]
            correct = y_pred.eq(batch_data[-1].data.view_as(y_pred)).long().cpu().sum()
            num_correct += correct.item()
            total_loss += batch_mean_loss * len(batch_data[-1])
        return {'accuracy': 1.0 * num_correct / len(dataset), 'loss': total_loss / len(dataset)}

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory)



from flgo.benchmark.toolkits.series.time_series_forecasting import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator
import os
import flgo
from flgo.benchmark.toolkits.series.time_series_forecasting.datasets import BuiltinClassDataset
from flgo.benchmark.toolkits import download_from_url, extract_from_gz, normalized
import numpy as np
import ujson as json

path = os.path.join(flgo.benchmark.path,'RAW_DATA', 'EXCHANGE_RATE')

class ExchangeRate(BuiltinClassDataset):
    def __init__(self, root, train=True, window=24, horizon=12, normalize=2):
        self.normalize = normalize
        self.url = \
            "https://github.com/laiguokun/multivariate-time-series-data/raw/master/exchange_rate/exchange_rate.txt.gz"
        self.file = 'w_' + str(window) + '_h_' + str(horizon) + '_'
        self.window = window
        self.horizon = horizon
        super(ExchangeRate, self).__init__(root, train)

    def download(self):
        os.makedirs(self.raw_folder, exist_ok=True)
        os.makedirs(self.processed_folder, exist_ok=True)
        src_path = os.path.join(self.raw_folder, 'exchange_rate.txt.gz')
        if not os.path.exists(src_path):
            download_from_url(self.url, src_path)
        src_path = extract_from_gz(src_path, os.path.join(self.raw_folder, 'exchange_rate.txt'))
        raw_data = np.loadtxt(src_path, delimiter=',')
        data = normalized(raw_data, self.normalize)
        self.split(data)

    def split(self, data):
        n, m = data.shape
        s1 = int(n * 0.8)
        train_idx = range(self.window + self.horizon - 1, s1)
        test_idx = range(s1, n)
        train_data = self.batchify(train_idx, data)
        with open(os.path.join(self.processed_folder,
                               'w_' + str(self.window) + '_h_' + str(self.horizon) + '_train_data.json'), 'w') as f:
            json.dump(train_data, f)
        test_data = self.batchify(test_idx, data)
        with open(os.path.join(self.processed_folder,
                               'w_' + str(self.window) + '_h_' + str(self.horizon) + '_test_data.json'), 'w') as f:
            json.dump(test_data, f)

    def batchify(self, idx_set, data):
        n = len(idx_set)
        X = []
        Y = []
        for i in range(n):
            end = idx_set[i] - self.horizon + 1
            start = end - self.window
            X.append(data[start:end, :].tolist())
            Y.append(data[idx_set[i], :].tolist())
        return {'x': X, 'y': Y}

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=path):
        super(TaskGenerator, self).__init__(os.path.split(os.path.dirname(__file__))[-1], rawdata_path, ExchangeRate)

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, ExchangeRate)

TaskCalculator = GeneralCalculator


from __future__ import division

import random
import tqdm
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import sys, os
from flgo.utils.fmodule import FModule
import torchvision.transforms as transforms

class Model(FModule):
    CLASSES = (
        "aeroplane",
        "bicycle",
        "bird",
        "boat",
        "bottle",
        "bus",
        "car",
        "cat",
        "chair",
        "cow",
        "diningtable",
        "dog",
        "horse",
        "motorbike",
        "person",
        "pottedplant",
        "sheep",
        "sofa",
        "train",
        "tvmonitor",
    )
    def __init__(self, cfgfile=os.path.join('/'.join(__file__.split('/')[:-1]),'yolov3.cfg')):
        super(Model, self).__init__()
        self.model = Darknet(cfgfile)
        self.model.apply(weights_init_normal)
        self.img_size = 384
        self.min_size = self.img_size - 3 * 32
        self.max_size = self.img_size + 3 * 32
        self.resize = transforms.Resize
        self.batch_count = 0
        self.size_interval = 10
        self.iou_thres=0.5
        self.conf_thres=0.5
        self.nms_thres=0.5

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def transform(self, images, targets, device=None):
        img_sizes = [tuple(img.shape[1:]) for img in images]
        # images = torch.stack([self.resize(img) for img in images])
        new_imgs = []
        labels = []
        for aid, ann in enumerate(targets):
            img = images[aid]
            boxes = []
            for obj in ann['annotation']['object']:
                box = (int(obj['bndbox']['xmin']), int(obj['bndbox']['xmax']), int(obj['bndbox']['ymin']), int(obj['bndbox']['ymax']))
                new_box = convert(img_sizes[aid], box)
                box = [self.CLASSES.index(obj['name'])]
                box.extend(new_box)
                boxes.append(box)
            img, img_targets = YoloTransforms(img, torch.FloatTensor(boxes), img_sizes[aid][0], img_sizes[aid][1])
            img_targets = torch.concat([torch.ones((img_targets.shape[0], 1))*aid, img_targets],dim=1)
            labels.append(img_targets)
            new_imgs.append(img)
        if self.batch_count%self.size_interval==0:
            self.img_size = random.choice(range(self.min_size, self.max_size + 1, 32))
        self.batch_count+=1
        images = torch.stack([self.resize((self.img_size, self.img_size))(img) for img in new_imgs])
        targets = torch.cat(labels)
        return images.to(device), targets.to(device)

    def organize_output(self, output):
        res = {}
        res['loss']=output[0]
        res['else']=output[1]
        return res

    def evaluate(self, dataloader=None):
        if dataloader is None or len(dataloader)==0: return {}
        self.model.eval()
        # Get dataloader
        labels = []
        losses = []
        sample_metrics = []  # List of tuples (TP, confs, pred)
        for batch_i, (imgs, targets) in enumerate(tqdm.tqdm(dataloader, desc="Detecting objects")):
            imgs, targets = self.transform(imgs, targets, self.get_device())
            labels += targets[:, 1].tolist()
            with torch.no_grad():
                loss, outputs = self.model(imgs, targets)
                outputs = non_max_suppression(outputs, conf_thres=self.conf_thres, nms_thres=self.nms_thres)
            targets[:, 2:] *= self.img_size
            losses.append(loss*len(imgs))
            sample_metrics += get_batch_statistics(outputs, targets, iou_threshold=self.iou_thres)
        # Concatenate sample statistics
        res = {'loss': sum(losses)/len(dataloader)}
        if len(sample_metrics)>0:
            true_positives, pred_scores, pred_labels = [np.concatenate(x, 0) for x in list(zip(*sample_metrics))]
            precision, recall, AP, f1, ap_class = ap_per_class(true_positives, pred_scores, pred_labels, labels)
            res.update({'precision':precision, 'recall':recall,'AP': AP, 'f1':f1, 'ap_class': ap_class})
        return res


def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)

def ap_per_class(tp, conf, pred_cls, target_cls):
    """ Compute the average precision, given the recall and precision curves.
    Source: https://github.com/rafaelpadilla/Object-Detection-Metrics.
    # Arguments
        tp:    True positives (list).
        conf:  Objectness value from 0-1 (list).
        pred_cls: Predicted object classes (list).
        target_cls: True object classes (list).
    # Returns
        The average precision as computed in py-faster-rcnn.
    """

    # Sort by objectness
    i = np.argsort(-conf)
    tp, conf, pred_cls = tp[i], conf[i], pred_cls[i]

    # Find unique classes
    unique_classes = np.unique(target_cls)

    # Create Precision-Recall curve and compute AP for each class
    ap, p, r = [], [], []
    for c in tqdm.tqdm(unique_classes, desc="Computing AP"):
        i = pred_cls == c
        n_gt = (target_cls == c).sum()  # Number of ground truth objects
        n_p = i.sum()  # Number of predicted objects

        if n_p == 0 and n_gt == 0:
            continue
        elif n_p == 0 or n_gt == 0:
            ap.append(0)
            r.append(0)
            p.append(0)
        else:
            # Accumulate FPs and TPs
            fpc = (1 - tp[i]).cumsum()
            tpc = (tp[i]).cumsum()

            # Recall
            recall_curve = tpc / (n_gt + 1e-16)
            r.append(recall_curve[-1])

            # Precision
            precision_curve = tpc / (tpc + fpc)
            p.append(precision_curve[-1])

            # AP from recall-precision curve
            ap.append(compute_ap(recall_curve, precision_curve))

    # Compute F1 score (harmonic mean of precision and recall)
    p, r, ap = np.array(p), np.array(r), np.array(ap)
    f1 = 2 * p * r / (p + r + 1e-16)

    return p, r, ap, f1, unique_classes.astype("int32")

def compute_ap(recall, precision):
    """ Compute the average precision, given the recall and precision curves.
    Code originally from https://github.com/rbgirshick/py-faster-rcnn.

    # Arguments
        recall:    The recall curve (list).
        precision: The precision curve (list).
    # Returns
        The average precision as computed in py-faster-rcnn.
    """
    # correct AP calculation
    # first append sentinel values at the end
    mrec = np.concatenate(([0.0], recall, [1.0]))
    mpre = np.concatenate(([0.0], precision, [0.0]))

    # compute the precision envelope
    for i in range(mpre.size - 1, 0, -1):
        mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

    # to calculate area under PR curve, look for points
    # where X axis (recall) changes value
    i = np.where(mrec[1:] != mrec[:-1])[0]

    # and sum (\Delta recall) * prec
    ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap

def get_batch_statistics(outputs, targets, iou_threshold):
    """ Compute true positives, predicted scores and predicted labels per sample """
    batch_metrics = []
    for sample_i in range(len(outputs)):

        if outputs[sample_i] is None:
            continue

        output = outputs[sample_i]
        pred_boxes = output[:, :4]
        pred_scores = output[:, 4]
        pred_labels = output[:, -1]

        true_positives = np.zeros(pred_boxes.shape[0])

        annotations = targets[targets[:, 0] == sample_i][:, 1:]
        target_labels = annotations[:, 0] if len(annotations) else []
        if len(annotations):
            detected_boxes = []
            target_boxes = annotations[:, 1:]

            for pred_i, (pred_box, pred_label) in enumerate(zip(pred_boxes, pred_labels)):

                # If targets are found break
                if len(detected_boxes) == len(annotations):
                    break

                # Ignore if label is not one of the target labels
                if pred_label not in target_labels:
                    continue
                if target_boxes.device!=pred_box.device: target_boxes = target_boxes.to(pred_box.device)
                iou, box_index = bbox_iou(pred_box.unsqueeze(0), target_boxes).max(0)
                if iou >= iou_threshold and box_index not in detected_boxes:
                    true_positives[pred_i] = 1
                    detected_boxes += [box_index]
        batch_metrics.append([true_positives, pred_scores, pred_labels])
    return batch_metrics

def YoloTransforms(img, boxes, height, width):
    h_factor, w_factor = (height, width)
    # Pad to square resolution
    img, pad = pad_to_square(img, 0)
    _, padded_h, padded_w = img.shape
    # Extract coordinates for unpadded + unscaled image
    x1 = w_factor * (boxes[:, 1] - boxes[:, 3] / 2)
    y1 = h_factor * (boxes[:, 2] - boxes[:, 4] / 2)
    x2 = w_factor * (boxes[:, 1] + boxes[:, 3] / 2)
    y2 = h_factor * (boxes[:, 2] + boxes[:, 4] / 2)
    # Adjust for added padding
    x1 += pad[0]
    y1 += pad[2]
    x2 += pad[1]
    y2 += pad[3]
    # Returns (x, y, w, h)
    boxes[:, 1] = ((x1 + x2) / 2) / padded_w
    boxes[:, 2] = ((y1 + y2) / 2) / padded_h
    boxes[:, 3] *= w_factor / padded_w
    boxes[:, 4] *= h_factor / padded_h

    return img, boxes

def pad_to_square(img, pad_value):
    c, h, w = img.shape
    dim_diff = np.abs(h - w)
    # (upper / left) padding and (lower / right) padding
    pad1, pad2 = dim_diff // 2, dim_diff - dim_diff // 2
    # Determine padding
    pad = (0, 0, pad1, pad2) if h <= w else (pad1, pad2, 0, 0)
    # Add padding
    img = F.pad(img, pad, "constant", value=pad_value)
    return img, pad

def convert(size, box):
    dw = 1. / size[1]
    dh = 1. / size[0]
    x = (box[0] + box[1]) / 2.0
    y = (box[2] + box[3]) / 2.0
    w = abs(box[1] - box[0])
    h = abs(box[3] - box[2])
    x = x * dw
    w = w * dw
    y = y * dh
    h = h * dh
    return [x, y, w, h]


def weights_init_normal(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        torch.nn.init.normal_(m.weight.data, 0.0, 0.02)
    elif classname.find("BatchNorm2d") != -1:
        torch.nn.init.normal_(m.weight.data, 1.0, 0.02)
        torch.nn.init.constant_(m.bias.data, 0.0)


def parse_model_config(path):
    """
    Takes a configuration file
    Returns a list of blocks. Each blocks describes a block in the neural
    network to be built. Block is represented as a dictionary in the list
    Parses the yolo-v3 layer configuration file and returns module definitions
    """

    file = open(path, 'r')
    lines = file.read().split('\n')
    lines = [x for x in lines if x and not x.startswith('#')]  # get rid of the empty lines
    lines = [x.rstrip().lstrip() for x in lines]  # get rid of fringe whitespaces
    module_defs = []
    for line in lines:
        if line.startswith('['):  # This marks the start of a new block
            module_defs.append({})
            module_defs[-1]['type'] = line[1:-1].rstrip()
            if module_defs[-1]['type'] == 'convolutional':
                module_defs[-1]['batch_normalize'] = 0
        else:
            key, value = line.split("=")
            value = value.strip()
            module_defs[-1][key.rstrip()] = value.strip()

    return module_defs

def xywh2xyxy(x):
    y = x.new(x.shape)
    y[..., 0] = x[..., 0] - x[..., 2] / 2
    y[..., 1] = x[..., 1] - x[..., 3] / 2
    y[..., 2] = x[..., 0] + x[..., 2] / 2
    y[..., 3] = x[..., 1] + x[..., 3] / 2
    return y

def non_max_suppression(prediction, conf_thres=0.5, nms_thres=0.4):
    """
    Removes detections with lower object confidence score than 'conf_thres' and performs
    Non-Maximum Suppression to further filter detections.
    Returns detections with shape:
        (x1, y1, x2, y2, object_conf, class_score, class_pred)
    """

    # From (center x, center y, width, height) to (x1, y1, x2, y2)
    prediction[..., :4] = xywh2xyxy(prediction[..., :4])
    output = [None for _ in range(len(prediction))]
    for image_i, image_pred in enumerate(prediction):
        # Filter out confidence scores below threshold
        image_pred = image_pred[image_pred[:, 4] >= conf_thres]
        # If none are remaining => process next image
        if not image_pred.size(0):
            continue
        # Object confidence times class confidence
        score = image_pred[:, 4] * image_pred[:, 5:].max(1)[0]
        # Sort by it
        image_pred = image_pred[(-score).argsort()]
        class_confs, class_preds = image_pred[:, 5:].max(1, keepdim=True)
        detections = torch.cat((image_pred[:, :5], class_confs.float(), class_preds.float()), 1)
        # Perform non-maximum suppression
        keep_boxes = []
        while detections.size(0):
            large_overlap = bbox_iou(detections[0, :4].unsqueeze(0), detections[:, :4]) > nms_thres
            label_match = detections[0, -1] == detections[:, -1]
            # Indices of boxes with lower confidence scores, large IOUs and matching labels
            invalid = large_overlap & label_match
            weights = detections[invalid, 4:5]
            # Merge overlapping bboxes by order of confidence
            detections[0, :4] = (weights * detections[invalid, :4]).sum(0) / weights.sum()
            keep_boxes += [detections[0]]
            detections = detections[~invalid]
        if keep_boxes:
            output[image_i] = torch.stack(keep_boxes)

    return output

def to_cpu(tensor):
    return tensor.detach().cpu()

def bbox_wh_iou(wh1, wh2):
    wh2 = wh2.t()
    w1, h1 = wh1[0], wh1[1]
    w2, h2 = wh2[0], wh2[1]
    inter_area = torch.min(w1, w2) * torch.min(h1, h2)
    union_area = (w1 * h1 + 1e-16) + w2 * h2 - inter_area
    return inter_area / union_area


def bbox_iou(box1, box2, x1y1x2y2=True):
    """
    Returns the IoU of two bounding boxes
    """
    if not x1y1x2y2:
        # Transform from center and width to exact coordinates
        b1_x1, b1_x2 = box1[:, 0] - box1[:, 2] / 2, box1[:, 0] + box1[:, 2] / 2
        b1_y1, b1_y2 = box1[:, 1] - box1[:, 3] / 2, box1[:, 1] + box1[:, 3] / 2
        b2_x1, b2_x2 = box2[:, 0] - box2[:, 2] / 2, box2[:, 0] + box2[:, 2] / 2
        b2_y1, b2_y2 = box2[:, 1] - box2[:, 3] / 2, box2[:, 1] + box2[:, 3] / 2
    else:
        # Get the coordinates of bounding boxes
        b1_x1, b1_y1, b1_x2, b1_y2 = box1[:, 0], box1[:, 1], box1[:, 2], box1[:, 3]
        b2_x1, b2_y1, b2_x2, b2_y2 = box2[:, 0], box2[:, 1], box2[:, 2], box2[:, 3]

    # get the corrdinates of the intersection rectangle
    inter_rect_x1 = torch.max(b1_x1, b2_x1)
    inter_rect_y1 = torch.max(b1_y1, b2_y1)
    inter_rect_x2 = torch.min(b1_x2, b2_x2)
    inter_rect_y2 = torch.min(b1_y2, b2_y2)
    # Intersection area
    inter_area = torch.clamp(inter_rect_x2 - inter_rect_x1 + 1, min=0) * torch.clamp(
        inter_rect_y2 - inter_rect_y1 + 1, min=0
    )
    # Union Area
    b1_area = (b1_x2 - b1_x1 + 1) * (b1_y2 - b1_y1 + 1)
    b2_area = (b2_x2 - b2_x1 + 1) * (b2_y2 - b2_y1 + 1)

    iou = inter_area / (b1_area + b2_area - inter_area + 1e-16)

    return iou



def build_targets(pred_boxes, pred_cls, target, anchors, ignore_thres):

    #ByteTensor = torch.cuda.ByteTensor if pred_boxes.is_cuda else torch.ByteTensor

    dev = pred_boxes.device
    nB = pred_boxes.size(0)
    nA = pred_boxes.size(1)
    nC = pred_cls.size(-1)
    nG = pred_boxes.size(2)

    # Output tensors
    obj_mask = torch.BoolTensor(nB, nA, nG, nG).fill_(0).to(dev)
    noobj_mask = torch.BoolTensor(nB, nA, nG, nG).fill_(1).to(dev)
    class_mask = torch.FloatTensor(nB, nA, nG, nG).fill_(0).to(dev)
    iou_scores = torch.FloatTensor(nB, nA, nG, nG).fill_(0).to(dev)
    tx = torch.FloatTensor(nB, nA, nG, nG).fill_(0).to(dev)
    ty = torch.FloatTensor(nB, nA, nG, nG).fill_(0).to(dev)
    tw = torch.FloatTensor(nB, nA, nG, nG).fill_(0).to(dev)
    th = torch.FloatTensor(nB, nA, nG, nG).fill_(0).to(dev)
    tcls = torch.FloatTensor(nB, nA, nG, nG, nC).fill_(0).to(dev)

    # Convert to position relative to box
    target_boxes = target[:, 2:6] * nG
    gxy = target_boxes[:, :2]
    gwh = target_boxes[:, 2:]
    # Get anchors with best iou
    ious = torch.stack([bbox_wh_iou(anchor, gwh) for anchor in anchors])
    best_ious, best_n = ious.max(0)
    # Separate target values
    b, target_labels = target[:, :2].long().t()
    gx, gy = gxy.t()
    gw, gh = gwh.t()
    gi, gj = gxy.long().t()
    # Set masks
    obj_mask[b, best_n, gj, gi] = 1
    noobj_mask[b, best_n, gj, gi] = 0

    # Set noobj mask to zero where iou exceeds ignore threshold
    for i, anchor_ious in enumerate(ious.t()):
        noobj_mask[b[i], anchor_ious > ignore_thres, gj[i], gi[i]] = 0

    # Coordinates
    tx[b, best_n, gj, gi] = gx - gx.floor()
    ty[b, best_n, gj, gi] = gy - gy.floor()
    # Width and height
    tw[b, best_n, gj, gi] = torch.log(gw / anchors[best_n][:, 0] + 1e-16)
    th[b, best_n, gj, gi] = torch.log(gh / anchors[best_n][:, 1] + 1e-16)
    # One-hot encoding of label
    tcls[b, best_n, gj, gi, target_labels] = 1
    # Compute label correctness and iou at best anchor
    class_mask[b, best_n, gj, gi] = (pred_cls[b, best_n, gj, gi].argmax(-1) == target_labels).float()
    iou_scores[b, best_n, gj, gi] = bbox_iou(pred_boxes[b, best_n, gj, gi], target_boxes, x1y1x2y2=False)

    tconf = obj_mask.float()
    return iou_scores, class_mask, obj_mask, noobj_mask, tx, ty, tw, th, tcls, tconf

sys.path.append(os.getcwd())

def create_modules(module_defs):
    """
    Constructs module list of layer blocks from module configuration in module_defs
    """
    hyperparams = module_defs.pop(0)
    output_filters = [int(hyperparams["channels"])]
    module_list = nn.ModuleList()
    for module_i, module_def in enumerate(module_defs):
        modules = nn.Sequential()

        if module_def["type"] == "convolutional":
            bn = int(module_def["batch_normalize"])
            filters = int(module_def["filters"])
            kernel_size = int(module_def["size"])
            pad = (kernel_size - 1) // 2
            modules.add_module(
                f"conv_{module_i}",
                nn.Conv2d(
                    in_channels=output_filters[-1],
                    out_channels=filters,
                    kernel_size=kernel_size,
                    stride=int(module_def["stride"]),
                    padding=pad,
                    bias=not bn,
                ),
            )
            if bn:
                modules.add_module(f"batch_norm_{module_i}", nn.BatchNorm2d(filters, momentum=0.9, eps=1e-5))
            if module_def["activation"] == "leaky":
                modules.add_module(f"leaky_{module_i}", nn.LeakyReLU(0.1))

        elif module_def["type"] == "maxpool":
            kernel_size = int(module_def["size"])
            stride = int(module_def["stride"])
            if kernel_size == 2 and stride == 1:
                modules.add_module(f"_debug_padding_{module_i}", nn.ZeroPad2d((0, 1, 0, 1)))
            maxpool = nn.MaxPool2d(kernel_size=kernel_size, stride=stride, padding=int((kernel_size - 1) // 2))
            modules.add_module(f"maxpool_{module_i}", maxpool)

        elif module_def["type"] == "upsample":
            upsample = Upsample(scale_factor=int(module_def["stride"]), mode="nearest")
            modules.add_module(f"upsample_{module_i}", upsample)

        elif module_def["type"] == "route":
            layers = [int(x) for x in module_def["layers"].split(",")]
            filters = sum([output_filters[1:][i] for i in layers])
            modules.add_module(f"route_{module_i}", EmptyLayer())

        elif module_def["type"] == "shortcut":
            filters = output_filters[1:][int(module_def["from"])]
            modules.add_module(f"shortcut_{module_i}", EmptyLayer())

        elif module_def["type"] == "yolo":
            anchor_idxs = [int(x) for x in module_def["mask"].split(",")]
            # Extract anchors
            anchors = [int(x) for x in module_def["anchors"].split(",")]
            anchors = [(anchors[i], anchors[i + 1]) for i in range(0, len(anchors), 2)]
            anchors = [anchors[i] for i in anchor_idxs]
            num_classes = int(module_def["classes"])
            img_size = int(hyperparams["height"])
            # Define detection layer
            yolo_layer = YOLOLayer(anchors, num_classes, img_size)
            modules.add_module(f"yolo_{module_i}", yolo_layer)
        # Register module list and number of output filters
        module_list.append(modules)
        output_filters.append(filters)

    return hyperparams, module_list

class Upsample(nn.Module):
    """ nn.Upsample is deprecated """

    def __init__(self, scale_factor, mode="nearest"):
        super(Upsample, self).__init__()
        self.scale_factor = scale_factor
        self.mode = mode

    def forward(self, x):
        x = F.interpolate(x, scale_factor=self.scale_factor, mode=self.mode)
        return x


class EmptyLayer(nn.Module):
    """Placeholder for 'route' and 'shortcut' layers"""

    def __init__(self):
        super(EmptyLayer, self).__init__()


class YOLOLayer(nn.Module):
    """Detection layer"""

    def __init__(self, anchors, num_classes, img_dim=416):
        super(YOLOLayer, self).__init__()
        self.anchors = anchors
        self.num_anchors = len(anchors)
        self.num_classes = num_classes
        self.ignore_thres = 0.5
        self.mse_loss = nn.MSELoss()
        self.bce_loss = nn.BCELoss()
        self.obj_scale = 1
        self.noobj_scale = 100
        self.metrics = {}
        self.img_dim = img_dim
        self.grid_size = 0  # grid size

    def compute_grid_offsets(self, grid_size, device=None):
        if device is None: device = torch.device('cpu')
        self.grid_size = grid_size
        g = self.grid_size
        self.stride = self.img_dim / self.grid_size
        # Calculate offsets for each grid
        self.grid_x = torch.arange(g).repeat(g, 1).view([1, 1, g, g]).type(torch.FloatTensor).to(device)
        self.grid_y = torch.arange(g).repeat(g, 1).t().view([1, 1, g, g]).type(torch.FloatTensor).to(device)
        self.scaled_anchors = torch.FloatTensor([(a_w / self.stride, a_h / self.stride) for a_w, a_h in self.anchors]).to(device)
        self.anchor_w = self.scaled_anchors[:, 0:1].view((1, self.num_anchors, 1, 1))
        self.anchor_h = self.scaled_anchors[:, 1:2].view((1, self.num_anchors, 1, 1))

    def forward(self, x, targets=None, img_dim=None):

        # Tensors for cuda support

        self.img_dim = img_dim
        num_samples = x.size(0)
        grid_size = x.size(2)

        prediction = (
            x.view(num_samples, self.num_anchors, self.num_classes + 5, grid_size, grid_size)
            .permute(0, 1, 3, 4, 2)
            .contiguous()
        )

        # Get outputs
        x = torch.sigmoid(prediction[..., 0])  # Center x
        y = torch.sigmoid(prediction[..., 1])  # Center y
        w = prediction[..., 2]  # Width
        h = prediction[..., 3]  # Height
        pred_conf = torch.sigmoid(prediction[..., 4])  # Conf
        pred_cls = torch.sigmoid(prediction[..., 5:])  # Cls pred.

        # If grid size does not match current we compute new offsets
        if grid_size != self.grid_size:
            self.compute_grid_offsets(grid_size, device=x.device)

        # Add offset and scale with anchors
        pred_boxes = torch.FloatTensor(prediction[..., :4].shape).to(x.device)
        pred_boxes[..., 0] = x.data + self.grid_x
        pred_boxes[..., 1] = y.data + self.grid_y
        pred_boxes[..., 2] = torch.exp(w.data) * self.anchor_w
        pred_boxes[..., 3] = torch.exp(h.data) * self.anchor_h

        output = torch.cat(
            (
                pred_boxes.view(num_samples, -1, 4) * self.stride,
                pred_conf.view(num_samples, -1, 1),
                pred_cls.view(num_samples, -1, self.num_classes),
            ),
            -1,
        )

        if targets is None:
            return output, 0
        else:
            iou_scores, class_mask, obj_mask, noobj_mask, tx, ty, tw, th, tcls, tconf = build_targets(
                pred_boxes=pred_boxes,
                pred_cls=pred_cls,
                target=targets,
                anchors=self.scaled_anchors,
                ignore_thres=self.ignore_thres,
            )

            # Loss : Mask outputs to ignore non-existing objects (except with conf. loss)
            loss_x = self.mse_loss(x[obj_mask], tx[obj_mask])
            loss_y = self.mse_loss(y[obj_mask], ty[obj_mask])
            loss_w = self.mse_loss(w[obj_mask], tw[obj_mask])
            loss_h = self.mse_loss(h[obj_mask], th[obj_mask])
            loss_conf_obj = self.bce_loss(pred_conf[obj_mask], tconf[obj_mask])
            loss_conf_noobj = self.bce_loss(pred_conf[noobj_mask], tconf[noobj_mask])
            loss_conf = self.obj_scale * loss_conf_obj + self.noobj_scale * loss_conf_noobj
            loss_cls = self.bce_loss(pred_cls[obj_mask], tcls[obj_mask])
            total_loss = loss_x + loss_y + loss_w + loss_h + loss_conf + loss_cls

            # Metrics
            cls_acc = 100 * class_mask[obj_mask].mean()
            conf_obj = pred_conf[obj_mask].mean()
            conf_noobj = pred_conf[noobj_mask].mean()
            conf50 = (pred_conf > 0.5).float()
            iou50 = (iou_scores > 0.5).float()
            iou75 = (iou_scores > 0.75).float()
            detected_mask = conf50 * class_mask * tconf
            precision = torch.sum(iou50 * detected_mask) / (conf50.sum() + 1e-16)
            recall50 = torch.sum(iou50 * detected_mask) / (obj_mask.sum() + 1e-16)
            recall75 = torch.sum(iou75 * detected_mask) / (obj_mask.sum() + 1e-16)

            self.metrics = {
                "loss": to_cpu(total_loss).item(),
                "x": to_cpu(loss_x).item(),
                "y": to_cpu(loss_y).item(),
                "w": to_cpu(loss_w).item(),
                "h": to_cpu(loss_h).item(),
                "conf": to_cpu(loss_conf).item(),
                "cls": to_cpu(loss_cls).item(),
                "cls_acc": to_cpu(cls_acc).item(),
                "recall50": to_cpu(recall50).item(),
                "recall75": to_cpu(recall75).item(),
                "precision": to_cpu(precision).item(),
                "conf_obj": to_cpu(conf_obj).item(),
                "conf_noobj": to_cpu(conf_noobj).item(),
                "grid_size": grid_size,
            }

            return output, total_loss


class Darknet(nn.Module):
    """YOLOv3 object detection model"""

    def __init__(self, config_path, img_size=416):
        super(Darknet, self).__init__()
        self.module_defs = parse_model_config(config_path)
        self.hyperparams, self.module_list = create_modules(self.module_defs)
        self.yolo_layers = [layer[0] for layer in self.module_list if hasattr(layer[0], "metrics")]
        self.img_size = img_size
        self.seen = 0
        self.header_info = np.array([0, 0, 0, self.seen, 0], dtype=np.int32)

    def forward(self, x, targets=None):
        img_dim = x.shape[2]
        loss = 0
        layer_outputs, yolo_outputs = [], []
        for i, (module_def, module) in enumerate(zip(self.module_defs, self.module_list)):
            if module_def["type"] in ["convolutional", "upsample", "maxpool"]:
                x = module(x)
            elif module_def["type"] == "route":
                x = torch.cat([layer_outputs[int(layer_i)] for layer_i in module_def["layers"].split(",")], 1)
            elif module_def["type"] == "shortcut":
                layer_i = int(module_def["from"])
                x = layer_outputs[-1] + layer_outputs[layer_i]
            elif module_def["type"] == "yolo":
                x, layer_loss = module[0](x, targets, img_dim)
                loss += layer_loss
                yolo_outputs.append(x)
            layer_outputs.append(x)
        yolo_outputs = to_cpu(torch.cat(yolo_outputs, 1))
        return yolo_outputs if targets is None else (loss, yolo_outputs)

    def load_darknet_weights(self, weights_path):
        """Parses and loads the weights stored in 'weights_path'"""

        # Open the weights file
        with open(weights_path, "rb") as f:
            header = np.fromfile(f, dtype=np.int32, count=5)  # First five are header values
            self.header_info = header  # Needed to write header when saving weights
            self.seen = header[3]  # number of images seen during training
            weights = np.fromfile(f, dtype=np.float32)  # The rest are weights

        # Establish cutoff for loading backbone weights
        cutoff = None
        if "darknet53.conv.74" in weights_path:
            cutoff = 75

        ptr = 0
        for i, (module_def, module) in enumerate(zip(self.module_defs, self.module_list)):
            if i == cutoff:
                break
            if module_def["type"] == "convolutional":
                conv_layer = module[0]
                if module_def["batch_normalize"]:
                    # Load BN bias, weights, running mean and running variance
                    bn_layer = module[1]
                    num_b = bn_layer.bias.numel()  # Number of biases
                    # Bias
                    bn_b = torch.from_numpy(weights[ptr : ptr + num_b]).view_as(bn_layer.bias)
                    bn_layer.bias.data.copy_(bn_b)
                    ptr += num_b
                    # Weight
                    bn_w = torch.from_numpy(weights[ptr : ptr + num_b]).view_as(bn_layer.weight)
                    bn_layer.weight.data.copy_(bn_w)
                    ptr += num_b
                    # Running Mean
                    bn_rm = torch.from_numpy(weights[ptr : ptr + num_b]).view_as(bn_layer.running_mean)
                    bn_layer.running_mean.data.copy_(bn_rm)
                    ptr += num_b
                    # Running Var
                    bn_rv = torch.from_numpy(weights[ptr : ptr + num_b]).view_as(bn_layer.running_var)
                    bn_layer.running_var.data.copy_(bn_rv)
                    ptr += num_b
                else:
                    # Load conv. bias
                    num_b = conv_layer.bias.numel()
                    conv_b = torch.from_numpy(weights[ptr : ptr + num_b]).view_as(conv_layer.bias)
                    conv_layer.bias.data.copy_(conv_b)
                    ptr += num_b
                # Load conv. weights
                num_w = conv_layer.weight.numel()
                conv_w = torch.from_numpy(weights[ptr : ptr + num_w]).view_as(conv_layer.weight)
                conv_layer.weight.data.copy_(conv_w)
                ptr += num_w

    def save_darknet_weights(self, path, cutoff=-1):
        """
            @:param path    - path of the new weights file
            @:param cutoff  - save layers between 0 and cutoff (cutoff = -1 -> all are saved)
        """
        fp = open(path, "wb")
        self.header_info[3] = self.seen
        self.header_info.tofile(fp)

        # Iterate through layers
        for i, (module_def, module) in enumerate(zip(self.module_defs[:cutoff], self.module_list[:cutoff])):
            if module_def["type"] == "convolutional":
                conv_layer = module[0]
                # If batch norm, load bn first
                if module_def["batch_normalize"]:
                    bn_layer = module[1]
                    bn_layer.bias.data.cpu().numpy().tofile(fp)
                    bn_layer.weight.data.cpu().numpy().tofile(fp)
                    bn_layer.running_mean.data.cpu().numpy().tofile(fp)
                    bn_layer.running_var.data.cpu().numpy().tofile(fp)
                # Load conv bias
                else:
                    conv_layer.bias.data.cpu().numpy().tofile(fp)
                # Load conv weights
                conv_layer.weight.data.cpu().numpy().tofile(fp)

        fp.close()


"""
This is a non-official implementation of 'Measuring the Effects
of Non-Identical Data Distribution for Federated Visual Classification'
(http://arxiv.org/abs/1909.06335).
"""

from .fedavg import Client
from .fedbase import BasicServer

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'beta': 0.01})
        self.v = self.model.zeros_like()

    def iterate(self):
        self.selected_clients = self.sample()
        models = self.communicate(self.selected_clients)['model']
        new_model = self.aggregate(models)
        self.v = self.beta*self.v + (self.model - new_model)
        self.model = self.model - self.v
        return


import zipfile

import numpy
import ujson as json
import os
import urllib
import numpy as np
import pandas as pd
import random

import torch
from torch.utils.data import Dataset


def download_from_url(url=None, filepath='.'):
    """Download dataset from url to filepath."""
    if url:
        urllib.request.urlretrieve(url, filepath)
    return filepath


def extract_from_zip(src_file, target_directory):
    """Unzip the .zip file (src_path) to target_path"""
    file = zipfile.ZipFile(src_file)
    file.extractall(target_directory)
    file.close()


def preprocessing(data, names, column_type, default_dic=None):
    res = [[] for i in range(len(data))]
    for index, type in enumerate(column_type):
        id = names[index]
        if type=='continuous':
            max_val = int(data[id].max())
            min_val = int(data[id].min())
            for i in range(len(data)):
                value = int(data[id][i])
                # data.loc[i, id] = (value - min_val) / (max_val - min_val)
                res[i].append((value - min_val) / (max_val - min_val))
        elif id == 'label':
            for i in range(len(data)):
                value = data[id][i]
                if value == '<=50K':
                    res[i].append(0)
                else:
                    res[i].append(1)
        else:
            if default_dic is None:
                keys = data[id].unique()
                key_dic = {}
                for i, k in enumerate(keys):
                    key_dic[k]=i
            else:
                key_dic = default_dic[index]
            # onehot
            for i in range(len(data)):
                value = data[id][i]
                # data.loc[i, id] = '0' * key_dic[value] + '1' + '0' * (len(key_dic) - key_dic[value] - 1)
                temp = [0] * key_dic[value] + [1] + [0] * (len(key_dic) - key_dic[value] - 1)
                res[i] = res[i] + temp
    return res

class BuiltinClassDataset(Dataset):

    def __init__(self, root, train=True):
        self.root = root
        self.train = train
        self.raw_folder = os.path.join(self.root, 'raw_data')
        self.processed_folder = os.path.join(self.root, 'processed_data')
        if not hasattr(self, 'file'):
            self.file = 'train_data.json' if self.train else 'test_data.json'
        else:
            self.file = self.file + 'train_data.json' if self.train else self.file + 'test_data.json'
        if not os.path.exists(os.path.join(self.processed_folder, self.file)):
            self.download()
        with open(os.path.join(self.processed_folder, self.file), 'r') as f:
            data = json.load(f)
        self.x = torch.tensor(data['x'], dtype=torch.float)
        self.y = torch.tensor(data['y'], dtype=torch.float)

    def __getitem__(self, index):
        return self.x[index], self.y[index]

    def __len__(self):
        return len(self.x)

    def download(self):
        raise NotImplementedError("please override the download() method in class {}".format(self.__class__.__name__))


class Adult(BuiltinClassDataset):
    """
    在download中将原始数据进行分片并划分为训练集和测试集，分别存储到train_data.json和test_data.json两个文件中
    文件路径默认设置为 self.processed_folder/XX_data.json
    """
    names = ['age', 'workclass', 'fnlwgt', 'education', 'education-num', 'marial-status', 'occupation', 'relationship',
             'race', 'sex', 'capital-gain', 'capital-loss', 'hours-per-week', 'native-country', 'label']
    column_type = ['continuous', 8, 'continuous', 16, 'continuous', 7, 14, 6, 5, 2, 'continuous', 'continuous',
                   'continuous', 42, 2]

    def __init__(self, root, train=True):
        self.url = {
            'train': "https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data",
            'test': "https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.test"
        }
        self.train = train
        self.file = 'adult_'
        self.root = root
        self.default_dic = [{} for i in range(len(Adult.names))]
        super().__init__(root, train)

    def get_dic(self, paths):
        data1 = pd.read_csv(paths[0], header=None, names=Adult.names, skipinitialspace=True)
        data2 = pd.read_csv(paths[1], header=None, names=Adult.names, skipinitialspace=True)
        data2.drop(index=[0], inplace=True)
        data2 = data2.reset_index(drop=True)
        for index, val in enumerate(Adult.column_type):
            if val == 'continuous':
                continue
            name = Adult.names[index]
            temp1 = data1[name].unique().tolist()
            temp2 = data2[name].unique().tolist()
            keys = list(set(temp1+temp2))
            self.default_dic[index] = {k: i for i, k in enumerate(keys)}


    def download(self):
        os.makedirs(self.raw_folder, exist_ok=True)
        os.makedirs(self.processed_folder, exist_ok=True)
        train_src_path = os.path.join(self.raw_folder, 'adult.data')
        test_src_path = os.path.join(self.raw_folder, 'adult.test')
        if not os.path.exists(train_src_path):
            download_from_url(self.url['train'], train_src_path)
        if not os.path.exists(test_src_path):
            download_from_url(self.url['test'], test_src_path)
        target_file = 'raw_data\\adult.data' if self.train is True else 'raw_data\\adult.test'
        paths = [os.path.join(self.root, 'raw_data\\adult.data'), os.path.join(self.root, 'raw_data\\adult.test')]
        self.get_dic(paths)
        self.to_json(os.path.join(self.root, target_file))

    def to_json(self, path):
        raw_data = pd.read_csv(path, header=None, names=Adult.names, skipinitialspace=True)
        if self.train is False:
            raw_data.drop(index=[0], inplace=True)
            raw_data = raw_data.reset_index(drop=True)
        processed_data = preprocessing(raw_data, Adult.names, Adult.column_type, self.default_dic)
        processed_data = np.array(processed_data)
        dits = {'x':processed_data[:, :-1].tolist(), 'y':processed_data[:, -1].tolist()}
        file_name = 'adult_train_data.json' if self.train is True else 'adult_test_data.json'
        with open(os.path.join(self.processed_folder, file_name), 'w') as f:
            json.dump(dits, f)



# import flgo
# pat = os.path.join(flgo.benchmark.path,'RAW_DATA', 'bankmarketing')

class BankMarketing(BuiltinClassDataset):
    """
    在download中将原始数据进行分片并划分为训练集和测试集，分别存储到train_data.json和test_data.json两个文件中
    文件路径默认设置为 self.processed_folder/XX_data.json
    """
    names = ["age", "job", "marital", "education", "default", "balance", "housing", "loan", "contact",
             "day", "month", "duration", "campaign", "pdays", "previous", "poutcome", "label"]
    column_type = ['continuous', 12, 3, 4, 2, 'continuous', 2, 2, 3, 'continuous', 12, 'continuous',
                   'continuous', 'continuous', 'continuous', 4, 2]

    def __init__(self, root, train=True):
        self.url = 'https://archive.ics.uci.edu/ml/machine-learning-databases/00222/bank.zip'
        self.train = train
        self.file = 'bankmarketing_'
        self.root = root
        super().__init__(root, train)


    def download(self):
        os.makedirs(self.raw_folder, exist_ok=True)
        os.makedirs(self.processed_folder, exist_ok=True)
        src_path = os.path.join(self.raw_folder, 'bank.zip')
        if not os.path.exists(src_path):
            download_from_url(self.url, src_path)
        target_file = os.path.join(self.root, 'raw_data\\bank-full.csv')
        if not os.path.exists(target_file):
            extract_from_zip(src_path, os.path.join(self.root, 'raw_data'))
        self.to_json(target_file)

    def to_json(self, path):
        raw_data = pd.read_csv(path, sep=';', header=0, names=BankMarketing.names)
        raw_data.loc[raw_data['pdays'] == -1, ['pdays']] = 99999999
        processed_data = preprocessing(raw_data, BankMarketing.names, BankMarketing.column_type)
        processed_data = np.array(processed_data)
        num = len(processed_data)
        index = [i for i in range(num)]
        train_num = int(num * 0.8)
        train = random.sample(index, train_num)
        test = [i for i in range(num) if i not in train]
        train_dic = {'x':processed_data[train, :-1].tolist(), 'y':processed_data[train, -1].tolist()}
        test_dic = {'x':processed_data[test, :-1].tolist(), 'y':processed_data[test, -1].tolist()}
        with open(os.path.join(self.processed_folder, 'bankmarketing_train_data.json'), 'w') as f:
            json.dump(train_dic, f)
        with open(os.path.join(self.processed_folder, 'bankmarketing_test_data.json'), 'w') as f:
            json.dump(test_dic, f)




import json
import urllib
import zipfile
import torch
from torch.utils.data import Dataset
import flgo.benchmark
import os.path
from flgo.benchmark.toolkits import BasicTaskGenerator, BasicTaskCalculator
from flgo.benchmark.base import XYHorizontalTaskPipe as TaskPipe, BasicTaskPipe
import collections
import re
import os
import os.path
import json
import csv
import numpy as np

def download_from_url(url= None, filepath = '.'):
    """Download dataset from url to filepath."""
    if url:
        urllib.request.urlretrieve(url, filepath)
    return filepath

def extract_from_zip(src_path, target_path):
    """Unzip the .zip file (src_path) to target_path"""
    f = zipfile.ZipFile(src_path)
    f.extractall(target_path)
    targets = f.namelist()
    f.close()
    return [os.path.join(target_path, tar) for tar in targets]

class SENTIMENT140(Dataset):
    def __init__(self, root, train=True, max_len=140):
        self.train = train
        self.root = root
        self.max_len = max_len

        if not os.path.exists(os.path.join(self.root, 'raw_data', 'embs.json')):
            self.load_emb()
        with open(os.path.join(self.root, 'raw_data', 'embs.json'), 'r') as inf:
            embs = json.load(inf)
        self.id2word = embs['vocab']
        self.word2id = {v: k for k, v in enumerate(self.id2word)}

        file = 'train_data.json' if self.train else 'test_data.json'
        if not os.path.exists(os.path.join(self.root, 'raw_data', file)):
            self.download()
        with open(os.path.join(self.root, 'raw_data', file), 'r') as f:
            data = json.load(f)
            if self.train:
                self.id = data['id']
            self.x = data['x']
            self.y = data['y']

    def __getitem__(self, index):
        return torch.tensor(self.x[index], dtype=torch.long), self.y[index]

    def __len__(self):
        return len(self.x)

    def load_emb(self):
        raw_path = os.path.join(self.root, 'raw_data')
        if not os.path.exists(raw_path):
            os.makedirs(raw_path)
        embs_path = download_from_url("https://nlp.stanford.edu/data/glove.6B.zip",
                                      os.path.join(raw_path, 'tmp'))
        tar_paths = extract_from_zip(embs_path, raw_path)
        os.remove(tar_paths[0])
        os.remove(tar_paths[1])
        os.remove(tar_paths[2])
        os.remove(embs_path)
        with open(tar_paths[3], 'r') as inf:
            lines = inf.readlines()
        lines = [i.split() for i in lines]
        vocab = [i[0] for i in lines]
        emb_floats = [[float(n) for n in i[1:]] for i in lines]
        emb_floats.append([0.0 for _ in range(300)])  # for unknown word
        js = {'vocab': vocab, 'emba': emb_floats}
        with open(os.path.join(raw_path, 'embs.json'), 'w') as ouf:
            json.dump(js, ouf)

    def download(self):
        raw_path = os.path.join(self.root, 'raw_data')
        if not os.path.exists(raw_path):
            os.makedirs(raw_path)
        src_path = download_from_url("https://cs.stanford.edu/people/alecmgo/trainingandtestdata.zip",
                                     os.path.join(raw_path, 'tmp'))
        tar_paths = extract_from_zip(src_path, raw_path)
        os.remove(src_path)
        with open(tar_paths[0], 'rt', encoding='ISO-8859-1') as f:
            reader = csv.reader(f)
            test = list(reader)
        with open(tar_paths[1], 'rt', encoding='ISO-8859-1') as f:
            reader = csv.reader(f)
            train = list(reader)

        test_user_data = {}
        for i in range(len(test)):
            y = 1 if test[i][0] == "4" else 0
            if test[i][4] not in test_user_data:
                test_user_data[test[i][4]] = {'x': [], 'y': []}
                test_user_data[test[i][4]]['x'].append(test[i][1:])
                test_user_data[test[i][4]]['y'].append(y)
            else:
                test_user_data[test[i][4]]['x'].append(test[i][1:])
                test_user_data[test[i][4]]['y'].append(y)
        test_users = list(test_user_data.keys())
        testXs = []
        testYs = []
        for user in test_users:
            examples = test_user_data[user]
            x = self.tokenizer(examples['x'])
            testXs.extend(x)
            testYs.extend(examples['y'])
        test_data = {
            'x': testXs,
            'y': testYs,
        }
        with open(os.path.join(self.root, 'raw_data', 'test_data.json'), 'w') as f:
            json.dump(test_data, f)

        train_user_data = {}
        for i in range(len(train)):
            y = 1 if train[i][0] == "4" else 0
            if train[i][4] not in train_user_data:
                train_user_data[train[i][4]] = {'x': [], 'y': []}
                train_user_data[train[i][4]]['x'].append(train[i][1:])
                train_user_data[train[i][4]]['y'].append(y)
            else:
                train_user_data[train[i][4]]['x'].append(train[i][1:])
                train_user_data[train[i][4]]['y'].append(y)
        train_users = list(train_user_data.keys())
        train_sample_ids = []
        trainXs = []
        trainYs = []
        cid = 0
        for user in train_users:
            examples = train_user_data[user]
            x = self.tokenizer(examples['x'])
            trainXs.extend(x)
            trainYs.extend(examples['y'])
            train_sample_ids.extend([cid] * len(x))
            cid += 1
        train_data = {
            'x': trainXs,
            'y': trainYs,
            'id': train_sample_ids
        }
        with open(os.path.join(self.root, 'raw_data', 'train_data.json'), 'w') as f:
            json.dump(train_data, f)
        os.remove(tar_paths[0])
        os.remove(tar_paths[1])

    def tokenizer(self, data):
        # [ID, Date, Query, User, Content]
        unk = len(self.word2id)
        processed_data = []
        for raw_text in data:
            line_list = self.split_line(raw_text[4])  # split phrase in words
            indl = [self.word2id[w] if w in self.word2id else unk for w in line_list[:self.max_len]]
            indl += [unk] * (self.max_len - len(indl))
            processed_data.append(indl)
        return processed_data

    def split_line(self, line):
        '''split given line/phrase into list of words
        Arguments:
            line: string representing phrase to be split
        Returnss:
            list of strings, with each string representing a word
        '''
        return re.findall(r"[\w']+|[.,!?;]", line)

class TaskGenerator(BasicTaskGenerator):
    def __init__(self,rawdata_path=os.path.join(flgo.benchmark.path,'RAW_DATA', 'SENTIMENT140')):
        super(TaskGenerator, self).__init__(benchmark='leaf_sent140',
                                            rawdata_path=rawdata_path)

    def load_data(self):
        self.train_data = SENTIMENT140(self.rawdata_path, train=True)
        self.test_data = SENTIMENT140(self.rawdata_path, train=False)
        return

    def partition(self):
        self.local_datas = self.partitioner(self.train_data)
        self.num_clients = len(self.local_datas)

class TaskPipe(BasicTaskPipe):
    class TaskDataset(torch.utils.data.Subset):
        def __init__(self, dataset, indices):
            super().__init__(dataset, indices)
            self.dataset = dataset
            self.indices = indices

        def __getitem__(self, idx):
            if isinstance(idx, list):
                return self.dataset[[self.indices[i] for i in idx]]
            return self.dataset[self.indices[idx]]

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'server_data': list(range(len(generator.test_data))),
                   'rawdata_path': generator.rawdata_path}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid], }
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        train_data = SENTIMENT140(root=self.feddata['rawdata_path'], train=True)
        test_data = SENTIMENT140(root=self.feddata['rawdata_path'], train=False)
        # rearrange data for server
        server_data_test, server_data_valid = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'valid': server_data_valid}}
        # rearrange data for clients
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'])
            cdata_train, cdata_valid = self.split_dataset(cdata, running_time_option['train_holdout'])
            task_data[cname] = {'train': cdata_train, 'valid': cdata_valid}
        return task_data

class TaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(TaskCalculator, self).__init__(device, optimizer_name)
        self.criterion = torch.nn.CrossEntropyLoss()
        self.DataLoader = torch.utils.data.DataLoader

    def compute_loss(self, model, data):
        """
        Args:
            model: the model to train
            data: the training dataset

        Returns:
            dict of train-one-step's result, which should at least contains the key 'loss'
        """
        tdata = self.to_device(data)
        outputs = model(tdata[0])
        loss = self.criterion(outputs, tdata[-1])
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0):
        """
        Metric = [mean_accuracy, mean_loss]
        Args:
            model:
            dataset:
            batch_size:

        Returns: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size==-1:batch_size=len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers)
        total_loss = 0.0
        num_correct = 0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data[0])
            batch_mean_loss = self.criterion(outputs, batch_data[-1]).item()
            y_pred = outputs.data.max(1, keepdim=True)[1]
            correct = y_pred.eq(batch_data[-1].data.view_as(y_pred)).long().cpu().sum()
            num_correct += correct.item()
            total_loss += batch_mean_loss * len(batch_data[-1])
        return {'accuracy': 1.0*num_correct/len(dataset), 'loss':total_loss/len(dataset)}

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers)


import torch
import torch.nn as nn
import torch.nn.functional as F
from flgo.utils.fmodule import FModule
from torch_geometric.nn import SAGEConv

class NeighborAggregator(nn.Module):
    def __init__(self, input_dim, output_dim, use_bias=False, aggr_method="mean"):
        super(NeighborAggregator, self).__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.use_bias = use_bias
        self.aggr_method = aggr_method
        self.weight = nn.Parameter(torch.Tensor(input_dim, output_dim))
        if self.use_bias:
            self.bias = nn.Parameter(torch.Tensor(self.output_dim))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.kaiming_uniform_(self.weight)
        if self.use_bias:
            nn.init.zeros_(self.bias)

    def forward(self, neighbor_feture):
        if self.aggr_method == "mean":
            aggr_neighbor = neighbor_feture.mean(dim=1)
        elif self.aggr_method == "sum":
            aggr_neighbor = neighbor_feture.sum(dim=1)
        elif self.aggr_method == "max":
            aggr_neighbor = neighbor_feture.max(dim=1)
        else:
            raise ValueError("Unknown aggr type, expected sum, max, or mean, but got {}".format(self.aggr_method))
        neighbor_hidden = torch.matmul(aggr_neighbor, self.weight)
        if self.use_bias:
            neighbor_hidden += self.bias

        return neighbor_hidden


class SageGCN(nn.Module):
    def __init__(self, input_dim, hidden_dim, activation=F.relu,
                 aggr_neighbor_method="mean", aggr_hidden_method="sum"):
        super(SageGCN, self).__init__()
        assert aggr_neighbor_method in ["mean", "sum", "max"]
        assert aggr_hidden_method in ["sum", "concat"]
        self.aggr_neighbor = aggr_neighbor_method
        self.aggr_hidden = aggr_hidden_method
        self.activation = activation
        self.aggregator = NeighborAggregator(input_dim, hidden_dim, aggr_method=aggr_neighbor_method)
        self.weight = nn.Parameter(torch.Tensor(input_dim, hidden_dim))

    def reset_parameters(self):
        nn.init.kaiming_uniform_(self.weight)

    def forward(self, src_node_features, neighbor_node_features):
        neighbor_hidden = self.aggregator(neighbor_node_features)
        self_hidden = torch.matmul(src_node_features, self.weight)

        if self.aggr_hidden == "sum":
            hidden = self_hidden + neighbor_hidden
        elif self.aggr_hidden == "concat":
            hidden = torch.cat([self_hidden, neighbor_hidden], dim=1)
        else:
            raise  ValueError("Expected sum or concat, got {}".format(self.aggr_hidden))
        if self.activation:
            return self.activation(hidden)
        else:
            return hidden


class Model(FModule):
    def __init__(self, hidden_dim=[64, 7], num_neighbors_list=[10, 10]):
        super(Model, self).__init__()
        self.input_dim = 1433
        self.num_neighbors_list = num_neighbors_list
        self.num_layers = len(num_neighbors_list)
        self.gcn = []
        self.gcn.append(SageGCN(1433, hidden_dim[0]))
        self.gcn.append(SageGCN(hidden_dim[0], hidden_dim[1], activation=None))

    def forward(self, node_features_list):
        hidden = node_features_list
        for l in range(self.num_layers):
            next_hidden = []
            gcn = self.gcn[l]
            for hop in range(self.num_layers - 1):
                src_node_features = hidden[hop]
                src_node_num = len(src_node_features)
                neighbor_node_features = hidden[hop + 1].\
                    view(src_node_num, self.num_neighbors_list[hop], -1)
                h = gcn(src_node_features, neighbor_node_features)
                next_hidden.append(h)
            hidden = next_hidden
        return hidden[0]

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


r"""
This module is to simulate arbitrary system heterogeneity that may occur in practice.
We conclude four types of system heterogeneity from existing works.
System Heterogeneity Description:
    1. **Availability**: the devices will be either available or unavailable at each moment, where only the
                    available devices can be selected to participate in training.

    2. **Responsiveness**: the responsiveness describes the length of the period from the server broadcasting the
                    gloabl model to the server receiving the locally trained model from a particular client.

    3. **Completeness**: since the server cannot fully control the behavior of devices,it's possible for devices to
                    upload imcomplete model updates (i.e. only training for a few steps).

    4. **Connectivity**: the clients who promise to complete training may suffer accidients so that the server may lose
                    connections with these client who will never return the currently trained local_movielens_recommendation model.

We build up a client state machine to simulate the four types of system heterogeneity, and provide high-level
APIs to allow customized system heterogeneity simulation.

**Example**: How to customize the system heterogeneity:
```python
>>> class MySimulator(flgo.simulator.base.BasicSimulator):
...     def update_client_availability(self):
...         # update the variable 'prob_available' and 'prob_unavailable' for all the clients
...         self.set_variable(self.all_clients, 'prob_available', [0.9 for _ in self.all_clients])
...         self.set_variable(self.all_clients, 'prob_unavailable', [0.1 for _ in self.all_clients])
...
...     def update_client_connectivity(self, client_ids):
...         # update the variable 'prob_drop' for clients in client_ids
...         self.set_variable(client_ids, 'prob_drop', [0.1 for _ in client_ids])
...
...     def update_client_responsiveness(self, client_ids, *args, **kwargs):
...         # update the variable 'latency' for clients in client_ids
...         self.set_variable(client_ids, 'latency', [np.random.randint(5,100) for _ in client_ids])
...
...     def update_client_completeness(self, client_ids, *args, **kwargs):
...         # update the variable 'working_amount' for clients in client_ids
...         self.set_variable(client_ids, 'working_amount',  [max(int(self.clients[cid].num_steps*np.random.rand()), 1) for cid in client_ids])
>>> r = flgo.init(task, algorithm=fedavg, Simulator=MySimulator)
>>> # The runner r will be runned under the customized system heterogeneity, where the clients' states will be flushed by
>>> # MySimulator.update_client_xxx at each moment of the virtual clock or particular events happen (i.e. a client was selected)
```

We also provide some preset Simulator like flgo.simulator.DefaultSimulator and flgo.simulator.
"""
from flgo.simulator.default_simulator import Simulator as DefaultSimulator
from flgo.simulator.phone_simulator import Simulator as PhoneSimulator



"""
This is a non-official implementation of the work 'Clustered Sampling:
Low-Variance and Improved Representativity for Clients Selection in
Federated Learning (https://arxiv.org/abs/2105.05883)'. The official
implementation is in https://github.com/Accenture//Labs-FederatedLearning/tree/clustered_sampling
"""

from .fedbase import BasicServer
from .fedavg import Client
import numpy as np
from flgo.utils import fmodule
import copy
from itertools import product
from scipy.cluster.hierarchy import linkage, fcluster

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        # m = self.clients_per_round, M = self.total_data_vol = sum(n_i), n_i = self.local_data_vols[i]
        self.init_algo_para({'alg':1, 'distance_type':'cos'})
        self.W = None
        zero_model = fmodule._model_to_tensor(self.model-self.model).cpu().numpy()
        self.update_history = [copy.deepcopy(zero_model) for _ in range(self.num_clients)]
        self.local_data_vols = [len(c.train_data) for c in self.clients]
        self.total_data_vol = sum(self.local_data_vols)

    def iterate(self):
        self.selected_clients = self.sample()
        # training
        models = self.communicate(self.selected_clients)['model']
        for model_k, cid in zip(models, self.received_clients):
            self.update_history[cid]=fmodule._model_to_tensor(model_k-self.model).cpu().numpy()
        # aggregate: pk = 1/K as default where K=len(selected_clients)
        self.model = self.aggregate(models)
        return len(models)>0

    def update_w(self, m, M, ns, alg=1):
        if alg==1:
            # clustering based on data size
            if self.W: return self.W
            # q: current_data_amount
            current_data_amount = 0
            # k: current_cluster_idx
            current_cluster_idx = 0
            clients = [cid for cid in range(self.num_clients)]
            r = [[0 for _ in range(self.num_clients)] for _ in range(m)]
            # order clients by descending importance of n_i in ns
            tmp = zip(ns, clients)
            tmp = sorted(tmp, key=lambda x: x[0], reverse=True)
            b_prior = 0
            for n_cid, cid in tmp:
                current_data_amount = current_data_amount + m * n_cid
                ai = current_data_amount // M
                bi = current_data_amount % M
                if ai > current_cluster_idx:
                    r[current_cluster_idx][cid] = M - b_prior
                    for l in range(current_cluster_idx + 1, ai):
                        r[l][cid] = M
                    if bi!=0:
                        r[ai][cid] = bi
                else:
                    r[ai][cid] = bi - b_prior
                current_cluster_idx = ai
                # update b_i-1
                b_prior = bi
            return [[1.0 * rki / M for rki in rk] for rk in r]
        elif self.alg==2:
            # clustering based on client similarity
            epsilon = int(10 ** 10)
            sim_matrix = np.zeros((self.num_clients, self.num_clients))
            for i, j in product(range(self.num_clients), range(self.num_clients)):
                sim_matrix[i, j] = self.get_similarity(self.update_history[i], self.update_history[j])
            linkage_matrix = linkage(sim_matrix, "ward")

            # associate each client to a cluster
            weights = [1.0 * ni / M for ni in ns]
            augmented_weights = [1.0*ni/M for ni in ns]

            for i in range(len(linkage_matrix)):
                idx_1, idx_2 = int(linkage_matrix[i, 0]), int(linkage_matrix[i, 1])

                new_weight = np.array(
                    [augmented_weights[idx_1] + augmented_weights[idx_2]]
                )
                augmented_weights = np.concatenate((augmented_weights, new_weight))
                linkage_matrix[i, 2] = int(new_weight * epsilon)

            clusters = fcluster(
                linkage_matrix, int(epsilon / m), criterion="distance"
            )

            n_clients, n_clusters = len(clusters), len(set(clusters))

            # calculate the data volumn of each cluster
            pop_clusters = np.zeros((n_clusters, 2)).astype(int)
            for i in range(n_clusters):
                pop_clusters[i, 0] = i + 1
                for client in np.where(clusters == i + 1)[0]:
                    pop_clusters[i, 1] += int(weights[client] * epsilon * m)
            # sort the clusters according to the data volumn
            pop_clusters = pop_clusters[pop_clusters[:, 1].argsort()]

            distri_clusters = np.zeros((m, n_clients)).astype(int)

            # m biggest clusters that will remain unchanged
            kept_clusters = pop_clusters[n_clusters - m:, 0]
            # allocate weights for the clients in the kept clusters
            for idx, cluster in enumerate(kept_clusters):
                for client in np.where(clusters == cluster)[0]:
                    distri_clusters[idx, client] = int(weights[client] * m * epsilon)
            k = 0
            for j in pop_clusters[: n_clusters - m, 0]:
                clients_in_j = np.where(clusters == j)[0]
                np.random.shuffle(clients_in_j)
                for client in clients_in_j:
                    weight_client = int(weights[client] * epsilon * m)
                    while weight_client > 0:
                        sum_proba_in_k = np.sum(distri_clusters[k])
                        u_i = min(epsilon - sum_proba_in_k, weight_client)
                        distri_clusters[k, client] = u_i
                        weight_client += -u_i
                        sum_proba_in_k = np.sum(distri_clusters[k])
                        if sum_proba_in_k == 1 * epsilon:
                            k += 1
            distri_clusters = distri_clusters.astype(float)
            for l in range(m):
                distri_clusters[l] /= np.sum(distri_clusters[l])
            return distri_clusters.tolist()

    def sample(self):
        self.W = self.update_w(self.clients_per_round, self.total_data_vol, self.local_data_vols, self.alg)
        all_clients = [cid for cid in range(self.num_clients)]
        selected_clients = []
        for k in range(self.clients_per_round):
            cid = np.random.choice(all_clients, 1, p=self.W[k])[0]
            selected_clients.append(cid)
        return list(selected_clients)

    def get_similarity(self, g1, g2):
        if self.distance_type == "L1":
            return np.sum(np.abs(g1-g2))
        elif self.distance_type == "L2":
            return np.sum((g1-g2)**2)
        elif self.distance_type == "cos":
            ng1 = np.sum(g1**2)
            ng2 = np.sum(g2**2)
            if ng1==0.0 or ng2==0.0:
                return 0.0
            else:
                return np.arccos(np.sum(g1*g2)/(np.sqrt(ng1*ng2)))


import torchvision
from flgo.benchmark.toolkits.cv.classification import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator
import flgo.benchmark
import os.path

transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor(), torchvision.transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))])
builtin_class = torchvision.datasets.SVHN
path = os.path.join(flgo.benchmark.path,'RAW_DATA', 'SVHN')
TaskCalculator = GeneralCalculator

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=path):
        super(TaskGenerator, self).__init__(os.path.split(os.path.dirname(__file__))[-1], rawdata_path, builtin_class, transform)
        self.train_additional_option = {'split':'train'}
        self.test_additional_option = {'split':'test'}

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, builtin_class, transform)


"""This is a non-official implementation of 'Federated Learning with Buffered Asynchronous Aggregation' (http://arxiv.org/abs/2106.06639). """
from .fedasync import Server as AsyncServer
from .fedbase import BasicClient
import copy

class Server(AsyncServer):
    def initialize(self):
        self.init_algo_para({'period': 1, 'k': 0, 'K': 10})
        self.tolerance_for_latency = 1000
        self.updated = True
        self.accumulate_delta = None

    def iterate(self):
        # Scheduler periodically triggers the idle clients to locally train the model
        self.selected_clients = self.sample() if (self.gv.clock.current_time % self.period) == 0 or self.gv.clock.current_time == 1 else []
        if len(self.selected_clients) > 0:
            self.gv.logger.info('Select clients {} at time {}'.format(self.selected_clients, self.gv.clock.current_time))
        # Record the timestamp of the selected clients
        #  for cid in self.selected_clients: self.client_taus[cid] = self.current_round
        # Check the currently received models
        res = self.communicate(self.selected_clients, asynchronous=True)
        received_models = res['model']
        received_client_ids = res['__cid']
        # if reveive client update
        if len(received_models) > 0:
            self.gv.logger.info('Receive new models from clients {} at time {}'.format(received_client_ids, self.gv.clock.current_time))
            flag = False
            if self.accumulate_delta == None: flag = True
            for id, model_k in enumerate(received_models):
                if flag == True:
                        self.accumulate_delta = model_k
                        flag = False
                else:
                    self.accumulate_delta += model_k
                self.k += 1
                if self.k == self.K:
                    self.accumulate_delta = self.accumulate_delta * pow(self.K, -1)
                    self.model = self.model - self.lr * self.accumulate_delta
                    self.accumulate_delta = None
                    flag = True
                    self.k = 0
            # update aggregation round and the flag `updated`
            self.current_round += 1
            self.updated = True
        return



class Client(BasicClient):

    def reply(self, svr_pkg):
        model = self.unpack(svr_pkg)
        model_0 = copy.deepcopy(model)
        self.train(model)
        model_q = model_0 - model
        cpkg = self.pack(model_q)
        return cpkg


from flgo.benchmark.mnist_classification.core import builtin_class, transform
from flgo.benchmark.base import BasicTaskPipe, BasicTaskCalculator
from flgo.benchmark.toolkits.cv.classification import BuiltinClassGenerator
import torch
import os.path
from torch.utils.data import Dataset
import flgo
import random
import json

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.path,'RAW_DATA', 'MNIST')):
        super(TaskGenerator, self).__init__('vertical_mnist_classification', rawdata_path, builtin_class, transform)

    def partition(self):
        self.local_datas = self.partitioner(self.train_data)
        self.num_objects = len(self.local_datas)
        if self.test_data is not None:
            self.test_local_datas = [{'sample_idxs':list(range(len(self.test_data))), 'pt_feature':pdata['pt_feature'], 'with_label':pdata['with_label']} for pdata in self.local_datas]

    def get_task_name(self):
        return '_'.join(['B-' + self.benchmark, 'P-' + str(self.partitioner), 'N-' + str(self.partitioner.num_parties)])

class PartialDataset(Dataset):
    def __init__(self, dataset, pt_feature=None, with_label=False, sample_idxs = []):
        self.dataset = dataset
        self.sample_idxs = sample_idxs
        self.pt_feature = pt_feature
        self.with_label = with_label

    def gen_id(self, *args, **kwargs) -> list:
        return list(range(len(self.dataset)))

    def partition(self, x):
        return (torch.split(x, self.pt_feature[1], dim=self.pt_feature[0]))[self.pt_feature[2]]

    def is_with_label(self):
        return self.with_label

    def __len__(self):
        return len(self.sample_idxs)

    def __getitem__(self, item):
        sidx = self.sample_idxs[item]
        x,y = self.dataset[sidx][0], self.dataset[sidx][1]
        sid = self.dataset.ids[sidx]
        if self.with_label:
            return self.partition(x), y, sid
        else:
            return self.partition(x), None, sid

    def get_batch_by_id(self, ids):
        try:
            xs = [self.partition(self.dataset[sid][0]) for sid in ids]
            ys = [self.dataset[sid][1] for sid in ids]
        except:
            raise ValueError("sample with id={} doesn't exists in the current dataset")
        return torch.stack(xs), torch.LongTensor(ys), ids

class TaskPipe(BasicTaskPipe):
    TaskDataset = PartialDataset
    def __init__(self, task_name):
        super().__init__(task_name)
        self.builtin_class = builtin_class
        self.transform = transform

    def save_task(self, generator):
        party_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'party_names': party_names,'rawdata_path': generator.rawdata_path, 'additional_option': generator.additional_option}
        for pid in range(len(party_names)):
            feddata[party_names[pid]] = {
                'data':{
                    'with_label': generator.local_datas[pid]['with_label'],
                    'pt_feature': generator.local_datas[pid]['pt_feature'],
                    'train':generator.local_datas[pid]['sample_idxs'],
                    'test': generator.test_local_datas[pid]['sample_idxs'],
                },
            }
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        train_data = self.builtin_class(root=self.feddata['rawdata_path'], download=True, train=True, transform=self.transform, **self.feddata['additional_option'])
        test_data = self.builtin_class(root=self.feddata['rawdata_path'], download=True, train=False, transform=self.transform, **self.feddata['additional_option'])
        train_data.ids = list(range(len(train_data)))
        test_data.ids = list(range(len(test_data)))
        train_sidxs = list(range(len(train_data)))
        random.shuffle(train_sidxs)
        valid_sample_idxs = train_sidxs[:int(len(train_data)* running_time_option['train_holdout'])]
        train_sample_idxs = train_sidxs[int(len(train_data)* running_time_option['train_holdout']):]
        task_data = {}
        for pid, party_name in enumerate(self.feddata['party_names']):
            pdata = self.feddata[party_name]['data']
            with_label, pt_feature, train_idxs, test_idxs = pdata['with_label'], pdata['pt_feature'], pdata['train'], pdata['test']
            local_train_data = self.TaskDataset(train_data, pt_feature, with_label, train_sample_idxs)
            local_valid_data = self.TaskDataset(train_data, pt_feature, with_label, valid_sample_idxs)
            local_test_data = self.TaskDataset(test_data, pt_feature, with_label, test_idxs)
            task_data[party_name] = {'train':local_train_data, 'valid':local_valid_data, 'test':local_test_data}
        return task_data

class TaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(TaskCalculator, self).__init__(device, optimizer_name)
        self.criterion = torch.nn.CrossEntropyLoss()
        self.DataLoader = torch.utils.data.DataLoader

    def compute_loss(self, model, data):
        """
        Args: model: the model to train
                 data: the training dataset
        Returns: dict of train-one-step's result, which should at least contains the key 'loss'
        """
        model.to(self.device)
        tdata = self.to_device(data)
        outputs = model(tdata[0])
        loss = self.criterion(outputs, tdata[-1])
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0):
        """
        Metric = [mean_accuracy, mean_loss]
        Args: model:
                 dataset:
                 batch_size:
        Returns: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size==-1:batch_size=len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers)
        total_loss = 0.0
        num_correct = 0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data[0])
            batch_mean_loss = self.criterion(outputs, batch_data[-1]).item()
            y_pred = outputs.data.max(1, keepdim=True)[1]
            correct = y_pred.eq(batch_data[-1].data.view_as(y_pred)).long().cpu().sum()
            num_correct += correct.item()
            total_loss += batch_mean_loss * len(batch_data[-1])
        return {'accuracy': 1.0*num_correct/len(dataset), 'loss':total_loss/len(dataset)}

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers)


import flgo.benchmark.toolkits.partition
import flgo.benchmark.exchange_rate_forecasting.model.GRU as gru

default_partitioner = flgo.benchmark.toolkits.partition.IIDPartitioner
default_partition_para = {'num_clients':100}
default_model = gru



from torch import nn
import torch.nn.functional as F
from flgo.utils.fmodule import FModule


class Model(FModule):
    def __init__(self):
        super(Model, self).__init__()
        self.fc = nn.Linear(108, 2, bias=True)

    def forward(self, x):
        return self.fc(x)

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


from torch.utils.data import Dataset
import torch
import os
import numpy as np

from flgo.benchmark.toolkits import download_from_url, extract_one_from_zip


class UCRArchiveDataset(Dataset):
    def __init__(self, root, dataset_name, train=True):
        self.root = root
        self.train = train
        self.raw_folder = os.path.join(self.root, 'raw_data')
        self.processed_folder = os.path.join(self.root, 'UCRArchive_2018')
        self.dataset_name = dataset_name
        self.file = '{}_TRAIN.tsv'.format(dataset_name) if self.train else '{}_TEST.tsv'.format(dataset_name)
        self.url = "https://www.cs.ucr.edu/~eamonn/time_series_data_2018/UCRArchive_2018.zip"
        if not os.path.exists(os.path.join(self.processed_folder, dataset_name, self.file)):
            self.download()
        with open(os.path.join(self.processed_folder, dataset_name, self.file), 'r') as f:
            data = np.loadtxt(f)
        self.x = torch.from_numpy(data[:, 1:]).to(torch.float)
        self.y = torch.from_numpy(data[:, 0]).to(torch.long) - 1

    def __getitem__(self, index):
        return self.x[index], self.y[index]

    def __len__(self):
        return len(self.x)

    def download(self):
        os.makedirs(self.raw_folder, exist_ok=True)
        src_path = os.path.join(self.raw_folder, 'UCRArchive_2018.zip')
        if not os.path.exists(src_path):
            download_from_url(self.url, src_path)
        extract_one_from_zip(src_path, self.root, file_name=os.path.join('UCRArchive_2018', self.dataset_name, self.file), pwd=b'someone')



import torch
import torch.nn.functional as F
from torch_geometric.nn import GCNConv
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self):
        super().__init__()
        self.conv1 = GCNConv(500, 16)
        self.conv2 = GCNConv(16,3)

    def forward(self, data):
        x, edge_index = data.x, data.edge_index
        x = self.conv1(x, edge_index)
        x = F.relu(x)
        x = F.dropout(x, training=self.training)
        x = self.conv2(x, edge_index)
        return F.log_softmax(x, dim=1)

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


from flgo.benchmark.enzymes_graph_classification.model import GIN
import flgo.benchmark.toolkits.visualization
import flgo.benchmark.toolkits.partition

default_partitioner = flgo.benchmark.toolkits.partition.IIDPartitioner
default_partition_para = {'num_clients':100}
visualize = flgo.benchmark.toolkits.visualization.visualize_by_class
default_model = GIN



import torchtext.transforms as T
from torch.hub import load_state_dict_from_url
import torchdata.datapipes.iter as dpiter
from torchtext.models import RobertaClassificationHead, XLMR_BASE_ENCODER
from tqdm import tqdm
import torchtext.functional as F
import torch
from torch.utils.data import DataLoader
from flgo.utils.fmodule import FModule
from torchtext.data.functional import to_map_style_dataset

padding_idx = 1
bos_idx = 0
eos_idx = 2
max_seq_len = 256
xlmr_vocab_path = r"https://download.pytorch.org/models/text/xlmr.vocab.pt"
xlmr_spm_model_path = r"https://download.pytorch.org/models/text/xlmr.sentencepiece.bpe.model"

text_transform = T.Sequential(
    T.SentencePieceTokenizer(xlmr_spm_model_path),
    T.VocabTransform(load_state_dict_from_url(xlmr_vocab_path)),
    T.Truncate(max_seq_len - 2),
    T.AddToken(token=bos_idx, begin=True),
    T.AddToken(token=eos_idx, begin=False),
)
num_classes = 2
input_dim = 768

def collect_fn(batch):
    # features = [d['token_ids'] for d in batch]
    input = F.to_tensor([d['token_ids'] for d in batch], padding_value=padding_idx)
    labels = torch.tensor([d["target"] for d in batch])
    return input, labels

def apply_transform(x):
    return text_transform(x[0]), x[1]

def feat2dict(x):
    return {'token_ids':x[0], 'target':x[1]}

def init_dataset(object):
    datanames = object.get_data_names()
    for key in datanames:
        data = object.get_data(key)
        if data is None: continue
        dp = dpiter.IterableWrapper(data)
        dp = dp.map(apply_transform)
        dp = dp.map(feat2dict)
        object.set_data(to_map_style_dataset(dp), key)
        object.calculator.set_collect_fn(collect_fn)

class Model(FModule):
    def __init__(self):
        super(Model, self).__init__()
        self.model = XLMR_BASE_ENCODER.get_model(head=RobertaClassificationHead(num_classes=num_classes, input_dim=input_dim))

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

def init_local_module(object):
    return

def init_global_module(object):
    if 'Server' in object.get_classname():
        model = Model().to(object.device)
        object.set_model(model)


import os

import torch
import torch.nn as nn
import json

from flgo.utils.fmodule import FModule

class Model(FModule):

    def __init__(self, embedding_dim=300, vocab_size=400001, hidden_size=256, output_dim=2):
        super(Model, self).__init__()
        with open(os.path.join('./benchmark/RAW_DATA/SENTIMENT140', 'raw_data', 'embs.json'), 'r') as inf:
            embs = json.load(inf)
        self.embeddings = nn.Embedding(num_embeddings=vocab_size, embedding_dim=embedding_dim, padding_idx=0)
        self.embeddings.weight.data.copy_(torch.Tensor(embs['emba']))
        self.embeddings.weight.requires_grad = False
        self.lstm = nn.LSTM(input_size=embedding_dim, hidden_size=hidden_size, num_layers=2, batch_first=True)
        self.fc = nn.Linear(hidden_size * 2, output_dim)

    def forward(self, input_seq):
        embeds = self.embeddings(input_seq.T)
        self.lstm.flatten_parameters()
        lstm_out, _ = self.lstm(embeds)
        encoding = torch.cat((lstm_out[0], lstm_out[-1]), dim=1)
        output = self.fc(encoding)
        return output

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import  torch_geometric.datasets
import flgo.benchmark
from flgo.benchmark.toolkits.graph.node_classification import *

path = os.path.join(flgo.benchmark.path,'RAW_DATA', 'PubMed')
builtin_class = torch_geometric.datasets.Planetoid

TaskCalculator = GeneralCalculator

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=path):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1],
                                            rawdata_path=rawdata_path, builtin_class=builtin_class, transform=None, pre_transform=None)
        self.additional_option = {'name': 'PubMed'}

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, buildin_class=builtin_class, transform=None)


from .utils.fflow import init, gen_task, gen_task_by_para, tune, run_in_parallel, multi_init_and_run

communicator = None

class VirtualCommunicator:
    """
    Communicator that simulates the communication phase between any two objects
    """
    def __init__(self, objects):
        self.objects_map = {obj.id:obj for obj in objects}
        self.objects = objects

    def request(self, source, target, package):
        # send package to the target object with `package` and `mtype`, and then listen from it
        return self.objects_map[target].message_handler(package)



"""
This is a non-official implementation of 'Tackling the Objective Inconsistency Problem
in Heterogeneous Federated Optimization' (http://arxiv.org/abs/2007.07481)
"""
from flgo.algorithm.fedbase import BasicServer, BasicClient
from flgo.utils import fmodule

class Server(BasicServer):
    def iterate(self):
        self.selected_clients = self.sample()
        # training
        res = self.communicate(self.selected_clients)
        models, taus = res['model'], res['tau']
        ds = [(model-self.model)/tauk for model, tauk in zip(models, taus)]
        local_data_vols = [c.datavol for c in self.clients]
        total_data_vol = sum(local_data_vols)
        self.model = self.aggregate(ds, taus, p = [1.0 * local_data_vols[cid] / total_data_vol for cid in self.received_clients])
        return

    def aggregate(self, ds, taus, p=[]):
        if not ds: return self.model
        if self.aggregation_option == 'weighted_scale':
            K = len(ds)
            N = self.num_clients
            tau_eff = sum([tauk*pk for tauk,pk in zip(taus, p)])
            delta = fmodule._model_sum([dk * pk for dk, pk in zip(ds, p)]) * N / K
        elif self.aggregation_option == 'uniform':
            tau_eff = 1.0*sum(taus)/len(ds)
            delta = fmodule._model_average(ds)

        elif self.aggregation_option == 'weighted_com':
            tau_eff = sum([tauk * pk for tauk, pk in zip(taus, p)])
            delta = fmodule._model_sum([dk * pk for dk, pk in zip(ds, p)])
        else:
            sump = sum(p)
            p = [pk/sump for pk in p]
            tau_eff = sum([tauk * pk for tauk, pk in zip(taus, p)])
            delta = fmodule._model_sum([dk * pk for dk, pk in zip(ds, p)])
        return self.model + tau_eff * delta

class Client(BasicClient):
    def pack(self, model):
        tau = self._working_amount if hasattr(self, '_working_amount') else self.num_steps
        return {
            "model" : model,
            "tau": tau,
        }


import torchvision.models
import flgo.utils.fmodule as fmodule

class Model(fmodule.FModule):
    def __init__(self):
        super(Model, self).__init__()
        self.model = torchvision.models.segmentation.fcn_resnet50(num_classes=21, aux_loss=True)

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


from __future__ import annotations

import random
from typing import Any

import torch.utils.data
import json
from flgo.benchmark.base import *
from torch.utils.data import DataLoader

def collate_fn(batch):
    enum_data = {k:[d[k] for d in batch] for k in batch[0]['__enum__']}
    unenum_data = {k:batch[0][k] for k in batch[0].keys() if k not in batch[0]['__enum__']}
    enum_data.update(unenum_data)
    return enum_data

class BuiltinClassGenerator(BasicTaskGenerator):
    def __init__(self, benchmark, rawdata_path, builtin_class, transform=None):
        super(BuiltinClassGenerator, self).__init__(benchmark, rawdata_path)
        self.builtin_class = builtin_class
        self.transform = transform
        self.additional_option = {}
        self.train_additional_option = {}
        self.test_additional_option = {}

    def load_data(self, *args, **kwargs):
        # load the datasets
        train_default_init_para = {'root': self.rawdata_path, 'download':True, 'train':True, 'transform':self.transform}
        test_default_init_para = {'root': self.rawdata_path, 'download':False, 'train':False, 'transform':self.transform}
        train_default_init_para.update(self.additional_option)
        train_default_init_para.update(self.train_additional_option)
        test_default_init_para.update(self.additional_option)
        test_default_init_para.update(self.test_additional_option)
        if 'kwargs' not in self.builtin_class.__init__.__annotations__:
            train_pop_key = [k for k in train_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
            test_pop_key = [k for k in test_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
            for k in train_pop_key: train_default_init_para.pop(k)
            for k in test_pop_key: test_default_init_para.pop(k)
        # init datasets
        self.train_data = self.builtin_class(**train_default_init_para)
        self.test_data = self.builtin_class(**test_default_init_para)
        # handle train_data to triple (user, item, rating)
        self.train_data = [tuple(self.train_data[idx][k] for k in self.train_data.enum_data) for idx in range(len(self.train_data))]

    def partition(self):
        self.local_datas = self.partitioner(self.train_data)
        self.num_clients = len(self.local_datas)

class BuiltinClassPipe(BasicTaskPipe):
    class TaskDataset(torch.utils.data.Dataset):
        def __init__(self, dataset, index):
            self.enum_data = dataset.enum_data
            self.unenum_data = dataset.unenum_data
            self.num_users = dataset.num_users
            self.num_items = dataset.num_items
            for attr in self.enum_data:
                setattr(self, attr, getattr(dataset, attr)[index])
            for attr in self.unenum_data:
                setattr(self, attr, getattr(dataset, attr))

        def __len__(self):
            if len(self.enum_data)>0:
                return len(getattr(self, self.enum_data[0]))
            else:
                return 0

        def __getitem__(self, item):
            data = {attr: getattr(self, attr)[item] for attr in self.enum_data}
            data.update({attr: getattr(self, attr) for attr in self.unenum_data})
            data.update({'__enum__': self.enum_data})
            return data

    def __init__(self, task_path, buildin_class, transform=None):
        """
        Args:
            task_path (str): the path of the task
            builtin_class (class): class in torchvision.datasets
            transform (torchvision.transforms.*): the transform
        """
        super(BuiltinClassPipe, self).__init__(task_path)
        self.builtin_class = buildin_class
        self.transform = transform

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'server_data': list(range(len(generator.test_data))),  'rawdata_path': generator.rawdata_path, 'additional_option': generator.additional_option, 'train_additional_option':generator.train_additional_option, 'test_additional_option':generator.test_additional_option, }
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid],}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        train_default_init_para = {'root': self.feddata['rawdata_path'], 'download': True, 'train': True,
                                   'transform': self.transform}
        test_default_init_para = {'root': self.feddata['rawdata_path'], 'download': True, 'train': False,
                                  'transform': self.transform}
        train_default_init_para.update(self.feddata['additional_option'])
        train_default_init_para.update(self.feddata['train_additional_option'])
        test_default_init_para.update(self.feddata['additional_option'])
        test_default_init_para.update(self.feddata['test_additional_option'])
        if 'kwargs' not in self.builtin_class.__init__.__annotations__:
            train_pop_key = [k for k in train_default_init_para.keys() if
                             k not in self.builtin_class.__init__.__annotations__]
            test_pop_key = [k for k in test_default_init_para.keys() if
                            k not in self.builtin_class.__init__.__annotations__]
            for k in train_pop_key: train_default_init_para.pop(k)
            for k in test_pop_key: test_default_init_para.pop(k)
        # init datasets
        self.train_data = self.builtin_class(**train_default_init_para)
        self.test_data = self.builtin_class(**test_default_init_para)
        test_idxs = list(range(len(self.test_data)))
        k = int(len(self.test_data)*(1-running_time_option['test_holdout']))
        server_test_idxs = test_idxs[:k]
        server_valid_idxs = test_idxs[k:]
        server_test_data = self.TaskDataset(self.test_data, server_test_idxs) if len(server_test_idxs)>0 else None
        server_valid_data = self.TaskDataset(self.test_data, server_valid_idxs) if len(server_valid_idxs)>0 else None
        task_data = {'server':{'test':server_test_data, 'valid':server_valid_data}}
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata_idxs = self.feddata[cname]['data']
            k1 = int(len(cdata_idxs)*(1-running_time_option['train_holdout']))
            k2 = int((1-0.5*running_time_option['train_holdout'])*len(cdata_idxs)) if running_time_option['local_test'] else int(len(cdata_idxs))
            cdata_train_idxs = cdata_idxs[:k1]
            cdata_valid_idxs = cdata_idxs[k1:k2]
            cdata_test_idxs = cdata_idxs[k2:]
            cdata_train = self.TaskDataset(self.train_data, cdata_train_idxs) if len(cdata_train_idxs)>0 else None
            cdata_valid = self.TaskDataset(self.train_data, cdata_valid_idxs) if len(cdata_valid_idxs)>0 else None
            cdata_test = self.TaskDataset(self.train_data, cdata_test_idxs) if len(cdata_test_idxs)>0 else None
            task_data[cname] = {'train':cdata_train, 'valid':cdata_valid, 'test':cdata_test}
        return task_data

class GeneralCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super().__init__(device, optimizer_name)
        self.criterion = torch.nn.L1Loss()
        self.DataLoader = DataLoader

    def compute_loss(self, model, data):
        batch_data = self.to_device(data)
        outputs = model(batch_data)
        loss = self.criterion(outputs,batch_data['rating'])
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        model.eval()
        if batch_size==-1:batch_size=len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)
        total_loss = 0.0
        mse_loss = 0.0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data)
            batch_mean_loss = self.criterion(outputs, batch_data['rating']).item()
            total_loss += batch_mean_loss * len(batch_data['rating'])
            mse_loss += ((outputs-batch_data['rating'])**2).sum().item()
        return {'mae':total_loss/len(dataset), 'rmse': np.sqrt(mse_loss/len(dataset))}

    def to_device(self, data):
        res = {}
        for k,v in data.items():
            try: res[k] = torch.Tensor(v).to(self.device)
            except: res[k] = v
        return res

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory, collate_fn=collate_fn)

class UserLevelCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super().__init__(device, optimizer_name)
        self.criterion = torch.nn.L1Loss()
        self.DataLoader = DataLoader

    def compute_loss(self, models:tuple|list, data:Any) -> dict:
        global_model, local_model = models
        if global_model is not None and hasattr(global_model, 'to'):global_model.to(self.device)
        if local_model is not None and hasattr(local_model, 'to'):local_model.to(self.device)
        batch_data = self.to_device(data)
        outputs = global_model(batch_data, local_model)
        loss = self.criterion(outputs, batch_data['rating'])
        return {'loss': loss}

    @torch.no_grad()
    def test(self, models:tuple|list, dataset:Any, batch_size=64, num_workers=0, pin_memory=False):
        global_model, local_model = models
        if global_model is not None and hasattr(global_model, 'to'):global_model.to(self.device)
        if local_model is not None and hasattr(local_model, 'to'):local_model.to(self.device)
        global_model.eval()
        if batch_size == -1: batch_size = len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers,
                                          pin_memory=pin_memory)
        total_loss = 0.0
        mse_loss = 0.0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = global_model(batch_data, local_model)
            batch_mean_loss = self.criterion(outputs, batch_data['rating']).item()
            total_loss += batch_mean_loss * len(batch_data['rating'])
            mse_loss += ((batch_data['rating']-outputs)**2).sum().cpu().item()
        return {'mae': total_loss / len(dataset), 'rmse': np.sqrt(mse_loss/len(dataset))}

    def to_device(self, data):
        return {k: torch.Tensor(v).to(self.device) for k, v in data.items()}

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers,
                          pin_memory=pin_memory, collate_fn=collate_fn)


import gzip

import numpy
import ujson as json
import os
import urllib
import numpy as np

import torch
from torch.utils.data import Dataset


def download_from_url(url=None, filepath='.'):
    """Download dataset from url to filepath."""
    if url:
        urllib.request.urlretrieve(url, filepath)
    return filepath


def extract_from_gz(src_file, target_file):
    """Unzip the .zip file (src_path) to target_path"""
    with open(target_file, 'wb') as f:
        zf = gzip.open(src_file, mode='rb')
        f.write(zf.read())
        zf.close()
    return target_file


def normalized(rawdata, normalize):
    n, m = rawdata.shape
    scale = numpy.ones(m)
    if normalize == 0:
        data = rawdata
    elif normalize == 1:
        data = rawdata / np.max(rawdata)
    elif normalize == 2:
        data = np.zeros((n, m))
        for i in range(m):
            scale[i] = np.max(np.abs(rawdata[:, i]))
            data[:, i] = rawdata[:, i] / np.max(np.abs(rawdata[:, i]))
    else:
        raise RuntimeError("The parameter 'normalize' can only take values from 0, 1, 2")
    return data


class BuiltinClassDataset(Dataset):

    def __init__(self, root, train=True):
        self.root = root
        self.train = train
        self.raw_folder = os.path.join(self.root, 'raw_data')
        self.processed_folder = os.path.join(self.root, 'processed_data')
        if not hasattr(self, 'file'):
            self.file = 'train_data.json' if self.train else 'test_data.json'
        else:
            self.file = self.file + 'train_data.json' if self.train else self.file + 'test_data.json'
        if not os.path.exists(os.path.join(self.processed_folder, self.file)):
            self.download()
        with open(os.path.join(self.processed_folder, self.file), 'r') as f:
            data = json.load(f)
        self.x = torch.tensor(data['x'], dtype=torch.float)
        self.y = torch.tensor(data['y'], dtype=torch.float)

    def __getitem__(self, index):
        return self.x[index], self.y[index]

    def __len__(self):
        return len(self.x)

    def download(self):
        raise NotImplementedError("please override the download() method in class {}".format(self.__class__.__name__))


class Electricity(BuiltinClassDataset):
    """
    在download中将原始数据进行分片并划分为训练集和测试集，分别存储到train_data.json和test_data.json两个文件中
    文件路径默认设置为 self.processed_folder/XX_data.json
    """

    def __init__(self, root, train=True, window=24, horizon=12, normalize=2):
        self.normalize = normalize
        self.url = \
            "https://github.com/laiguokun/multivariate-time-series-data/raw/master/electricity/electricity.txt.gz"
        self.file = 'w_' + str(window) + '_h_' + str(horizon) + '_'
        self.window = window
        self.horizon = horizon
        super(Electricity, self).__init__(root, train)

    def download(self):
        os.makedirs(self.raw_folder, exist_ok=True)
        os.makedirs(self.processed_folder, exist_ok=True)
        src_path = os.path.join(self.raw_folder, 'electricity.txt.gz')
        if not os.path.exists(src_path):
            download_from_url(self.url, src_path)
        src_path = extract_from_gz(src_path, os.path.join(self.raw_folder, 'electricity.txt'))
        raw_data = np.loadtxt(src_path, delimiter=',')
        data = normalized(raw_data, self.normalize)
        self.split(data)

    def split(self, data):
        n, m = data.shape
        s1 = int(n * 0.8)
        train_idx = range(self.window + self.horizon - 1, s1)
        test_idx = range(s1, n)
        train_data = self.batchify(train_idx, data)
        with open(os.path.join(self.processed_folder, 'w_' + str(self.window) + '_h_' + str(self.horizon) + '_train_data.json'), 'w') as f:
            json.dump(train_data, f)
        test_data = self.batchify(test_idx, data)
        with open(os.path.join(self.processed_folder, 'w_' + str(self.window) + '_h_' + str(self.horizon) + '_test_data.json'), 'w') as f:
            json.dump(test_data, f)

    def batchify(self, idx_set, data):
        n = len(idx_set)
        X = []
        Y = []
        for i in range(n):
            end = idx_set[i] - self.horizon + 1
            start = end - self.window
            X.append(data[start:end, :].tolist())
            Y.append(data[idx_set[i], :].tolist())
        return {'x': X, 'y': Y}





import sys
import numpy as np
from abc import ABCMeta, abstractmethod
import functools
import torch
import heapq

class PriorityQueue:
    r"""Priority Queue"""
    def __init__(self):
        self.queue = []

    def size(self):
        r"""The size of the queue"""
        return len(self.queue)

    def empty(self):
        r"""Return whether the queue is empty"""
        return len(self.queue)==0

    def put(self, item):
        r"""Put item into the queue"""
        heapq.heappush(self.queue, item)

    def get(self):
        r"""Get item from the queue"""
        return heapq.heappop(self.queue)

class AbstractSimulator(metaclass=ABCMeta):
    @abstractmethod
    def flush(self):
        # flush the states for all the things in the system as time steps
        pass

random_seed_gen = None
random_module = None

def seed_generator(seed=0):
    """Return an integer as the seed"""
    while True:
        yield seed+1
        seed+=1

def size_of_package(package):
    r"""
    Compute the size of the package

    Args:
        package (dict): the pacakge

    Returns:
        size (int): the size of the package
    """
    size = 0
    for v in package.values():
        if type(v) is torch.Tensor:
            size += sys.getsizeof(v.storage())
        else:
            size += v.__sizeof__()
    return size

class ElemClock:
    r"""Simulate the clock by the timestamp of each Element"""
    class Elem:
        r"""
        Element with a timestamp

        Args:
            x: element
            time (int): the timestamp
        """
        def __init__(self, x, time):
            self.x = x
            self.time = time

        def __str__(self):
            return '{} at Time {}'.format(self.x, self.time)

        def __lt__(self, other):
            return self.time < other.time

    def __init__(self):
        self.q = PriorityQueue()
        self.time = 0
        self.simulator = None

    def step(self, delta_t=1):
        r"""
        Step delta_t units of the virtual time

        Args:
            delta_t (int): the delta of time
        """
        if delta_t < 0: raise RuntimeError("Cannot inverse time of simulator.base.clock.")
        if self.simulator is not None:
            for t in range(delta_t):
                self.simulator.flush()
        self.time += delta_t

    def set_time(self, t):
        r"""
        Set time

        Args:
            t (int): time
        """
        if t < self.time: raise RuntimeError("Cannot inverse time of simulator.base.clock.")
        self.time = t

    def put(self, x, time):
        r"""
        Put an element into the time queue with timestamp

        Args:
            x: element
            time (int): the timestamp
        """
        self.q.put(self.Elem(x, time))

    def get(self):
        r"""
        Get an element from the queue

        Returns:
            the element in the nearest coming time
        """
        if self.q.empty(): return None
        return self.q.get().x

    def get_until(self, t):
        r"""
        Get elements from the queue until time t

        Args:
            t (int): time

        Returns:
            a list of elements whose timestamps is no larger than t
        """
        res = []
        while not self.empty():
            elem = self.q.get()
            if elem.time > t:
                self.put(elem.x, elem.time)
                break
            pkg = elem.x
            res.append(pkg)
        return res

    def get_sofar(self):
        r"""
        Get elements from the queue until now

        Returns:
            a list of elements whose timestamps is no larger than the current time
        """
        return self.get_until(self.current_time)

    def gets(self):
        r"""
        Get all the elements in the queue

        Returns:
            a list of elements in the queue
        """
        if self.empty(): return []
        res = []
        while not self.empty(): res.append(self.q.get())
        res = [rx.x for rx in res]
        return res

    def clear(self):
        r"""
        Clear the queue
        """
        while not self.empty():
            self.get()

    def conditionally_clear(self, f):
        r"""
        Clear elements if f(element) is False

        Args:
            f (function): a function that receives element and returns bool variable
        """
        buf = []
        while not self.empty(): buf.append(self.q.get())
        for elem in buf:
            if not f(elem.x): self.q.put(elem)
        return

    def empty(self):
        r"""Return whether the queue is empty"""
        return self.q.empty()

    @ property
    def current_time(self):
        r"""Return the current time"""
        return self.time

    def register_simulator(self, simulator):
        r"""Set self.simulator=simulator"""
        self.simulator = simulator

class BasicSimulator(AbstractSimulator):
    r"""
    Simulate the system heterogeneity with the client state machine.

    Args:
        object (list): a list of objects in the federated scenario
    """
    _STATE = ['offline', 'idle', 'selected', 'working', 'dropped']
    _VAR_NAMES = ['prob_available', 'prob_unavailable', 'prob_drop', 'working_amount', 'latency']
    def __init__(self, objects, *args, **kwargs):
        if len(objects)>0:
            self.server = objects[0]
            self.clients = objects[1:]
        else:
            self.server = None
            self.clients = []
        self.all_clients = list(range(len(self.clients)))
        self.random_module = np.random.RandomState(0)
        # client states and the variables
        self.client_states = ['idle' for _ in self.clients]
        self.roundwise_fixed_availability = False
        self.availability_latest_round = -1
        self.variables = [{
            'prob_available': 1.,
            'prob_unavailable': 0.,
            'prob_drop': 0.,
            'working_amount': c.num_steps,
            'latency': 0,
        } for c in self.clients]
        for var in self._VAR_NAMES:
            self.set_variable(self.all_clients, var, [self.variables[cid][var] for cid in self.all_clients])
        self.state_counter = [{'dropped_counter': 0, 'latency_counter': 0, } for _ in self.clients]

    def get_client_with_state(self, state='idle'):
        r"""
        Get clients according to their states.

        Args:
            state (str): the state in ['offline', 'idle', 'selected', 'working', 'dropped']

        Returns:
            a list of clients whose states are state
        """
        return [cid for cid, cstate in enumerate(self.client_states) if cstate == state]

    def set_client_state(self, client_ids, state):
        r"""
        Set the states of clients in client_ids to the state

        Args:
            client_ids (list): a list of clients' ids
            state (str): the state in ['offline', 'idle', 'selected', 'working', 'dropped']

        Returns:
            a list of clients whose states are state
        """
        if state not in self._STATE: raise RuntimeError('{} not in the default state'.format(state))
        if type(client_ids) is not list: client_ids = [client_ids]
        for cid in client_ids: self.client_states[cid] = state
        if state == 'dropped':
            self.set_client_dropped_counter(client_ids)
        if state == 'working':
            self.set_client_latency_counter(client_ids)
        if state == 'idle':
            self.reset_client_counter(client_ids)

    def set_client_latency_counter(self, client_ids = []):
        r"""Set the latency_counter"""
        if type(client_ids) is not list: client_ids = [client_ids]
        for cid in client_ids:
            self.state_counter[cid]['dropped_counter'] = 0
            self.state_counter[cid]['latency_counter'] = self.variables[cid]['latency']

    def set_client_dropped_counter(self, client_ids = []):
        r"""Set the dropped_counter"""
        if type(client_ids) is not list: client_ids = [client_ids]
        for cid in client_ids:
            self.state_counter[cid]['latency_counter'] = 0
            self.state_counter[cid]['dropped_counter'] = self.server.get_tolerance_for_latency()

    def reset_client_counter(self, client_ids = []):
        r"""Reset the clients' counter"""
        if type(client_ids) is not list: client_ids = [client_ids]
        for cid in client_ids:
            self.state_counter[cid]['dropped_counter'] = self.state_counter[cid]['latency_counter'] = 0
        return

    @property
    def idle_clients(self):
        """Return ideal clients"""
        return self.get_client_with_state('idle')

    @property
    def working_clients(self):
        """Return working clients"""
        return self.get_client_with_state('working')

    @property
    def offline_clients(self):
        """Return offline clients"""
        return self.get_client_with_state('offline')

    @property
    def selected_clients(self):
        """Return the selected clients"""
        return self.get_client_with_state('selected')

    @property
    def dropped_clients(self):
        """Return the dropped clients"""
        return self.get_client_with_state('dropped')

    def get_variable(self, client_ids, varname):
        r"""
        Get the simulator-private variables of the clients in client_ids according to varname

        Args:
            client_ids (list): a list of clients' ids
            varname (str): the name of the simulator-private variable

        Returns:
            the simulator-private variables of the clients in client_ids
        """
        if len(self.variables) ==0: return None
        if type(client_ids) is not list: client_ids = [client_ids]
        return [self.variables[cid][varname] if varname in self.variables[cid].keys() else None for cid in client_ids]

    def set_variable(self, client_ids, varname, values):
        r"""
        Set the simulator-private variables of the clients in client_ids to values

        Args:
            client_ids (list): a list of clients' ids
            varname (str): the name of the simulator-private variable
            values (list): a list of things
        """
        if type(client_ids) is not list: client_ids = [client_ids]
        if type(values) is not list: values = [values]
        assert len(client_ids) == len(values)
        for cid, v in zip(client_ids, values):
            self.variables[cid][varname] = v
            setattr(self.clients[cid], '_'+varname, v)

    def update_client_availability(self, *args, **kwargs):
        """API to update client availability every time unit"""
        return

    def update_client_connectivity(self, client_ids, *args, **kwargs):
        """API to update client connectivity every time unit"""
        return

    def update_client_completeness(self, client_ids, *args, **kwargs):
        """API to update client completeness every time unit"""
        return

    def update_client_responsiveness(self, client_ids, *args, **kwargs):
        """API to update client responsiveness every time unit"""
        return

    def flush(self):
        """Flush the client state machine as time goes by"""
        # +++++++++++++++++++ availability +++++++++++++++++++++
        # change self.variables[cid]['prob_available'] and self.variables[cid]['prob_unavailable'] for each client `cid`
        self.update_client_availability()
        # update states for offline & idle clients
        if len(self.idle_clients)==0 or not self.roundwise_fixed_availability or self.server.current_round > self.availability_latest_round:
            self.availability_latest_round = self.server.current_round
            offline_clients = {cid: 'offline' for cid in self.offline_clients}
            idle_clients = {cid:'idle' for cid in self.idle_clients}
            for cid in offline_clients:
                if (self.random_module.rand() <= self.variables[cid]['prob_available']): offline_clients[cid] = 'idle'
            for cid in self.idle_clients:
                if  (self.random_module.rand() <= self.variables[cid]['prob_unavailable']): idle_clients[cid] = 'offline'
            new_idle_clients = [cid for cid in offline_clients if offline_clients[cid] == 'idle']
            new_offline_clients = [cid for cid in idle_clients if idle_clients[cid] == 'offline']
            self.set_client_state(new_idle_clients, 'idle')
            self.set_client_state(new_offline_clients, 'offline')
        # update states for dropped clients
        for cid in self.dropped_clients:
            self.state_counter[cid]['dropped_counter'] -= 1
            if self.state_counter[cid]['dropped_counter'] < 0:
                self.state_counter[cid]['dropped_counter'] = 0
                self.client_states[cid] = 'offline'
                if (self.random_module.rand() < self.variables[cid]['prob_unavailable']):
                    self.set_client_state([cid], 'offline')
                else:
                    self.set_client_state([cid], 'idle')
        # Remark: the state transfer fo working clients is instead made once the server received from clients
        # # update states for working clients
        # for cid in self.working_clients:
        #     self.state_counter[cid]['latency_counter'] -= 1
        #     if self.state_counter[cid]['latency_counter'] < 0:
        #         self.state_counter[cid]['latency_counter'] = 0
        #         self.set_client_state([cid], 'offline')

#================================================Decorators==========================================
# Time Counter for any function which forces the `cfg.clock` to
# step one unit of time once the decorated function is called
# def time_step(f):
#     def f_timestep(*args, **kwargs):
#         cfg.clock.step()
#         return f(*args, **kwargs)
#     return f_timestep

# sampling phase
def with_availability(sample):
    r"""
    The decorator for sampling with client availability

    Example:
    ```python
        >>> import flgo.algorithm.fedbase
        >>> import flgo.simulator.base as ss
        >>> class Server(flgo.algorithm.fedbase.BasicServer):
        ...     @ss.with_availability
        ...     def sample(self):
        ...         ...
    ```
    """
    @functools.wraps(sample)
    def sample_with_availability(self):
        available_clients = self.gv.simulator.idle_clients
        # ensure that there is at least one client to be available at the current moment
        # while len(available_clients) == 0:
        #     self.gv.clock.step()
        #     available_clients = self.gv.simulator.idle_clients
        # call the original sampling function
        selected_clients = sample(self)
        # filter the selected but unavailable clients
        effective_clients = set(selected_clients).intersection(set(available_clients))
        # return the selected and available clients (e.g. sampling with replacement should be considered here)
        self._unavailable_selected_clients = [cid for cid in selected_clients if cid not in effective_clients]
        if len(self._unavailable_selected_clients)>0:
            self.gv.logger.info('The selected clients {} are not currently available.'.format(self._unavailable_selected_clients))
        selected_clients = [cid for cid in selected_clients if cid in effective_clients]
        self.gv.simulator.set_client_state(selected_clients, 'selected')
        return selected_clients
    return sample_with_availability

# communicating phase
def with_dropout(communicate):
    r"""
    The decorator for communicating to simulate the scene where clients may drop out

    Example:
    ```python
        >>> import flgo.algorithm.fedbase
        >>> import flgo.simulator.base as ss
        >>> class Server(flgo.algorithm.fedbase.BasicServer):
        ...     @ss.with_dropout
        ...     def communicate(self,...):
        ...         ...
    ```
    """
    @functools.wraps(communicate)
    def communicate_with_dropout(self, selected_clients, mtype=0, asynchronous=False):
        if len(selected_clients) > 0:
            self.gv.simulator.update_client_connectivity(selected_clients)
            probs_drop = self.gv.simulator.get_variable(selected_clients, 'prob_drop')
            self._dropped_selected_clients = [cid for cid,prob in zip(selected_clients, probs_drop) if self.gv.simulator.random_module.rand() <= prob]
            self.gv.simulator.set_client_state(self._dropped_selected_clients, 'dropped')
            return communicate(self, [cid for cid in selected_clients if cid not in self._dropped_selected_clients], mtype, asynchronous)
        else:
            return communicate(self, selected_clients, mtype, asynchronous)
    return communicate_with_dropout

# communicating phase
def with_latency(communicate_with):
    r"""
    The decorator to simulate the scene where there are network latencies during communication

    Example:
    ```python
        >>> import flgo.algorithm.fedbase
        >>> import flgo.simulator.base as ss
        >>> class Server(flgo.algorithm.fedbase.BasicServer):
        ...     @ss.with_latency
        ...     def communicate_with(self,...):
        ...         ...
    ```
    """
    @functools.wraps(communicate_with)
    def delayed_communicate_with(self, target_id, package):
        # Calculate latency for the target client
        # Set local_movielens_recommendation model size of clients for computation cost estimation
        model_size = package['model'].count_parameters(output=False) if 'model' in package.keys() else 0
        self.gv.simulator.set_variable(target_id, '__model_size', model_size)
        # Set downloading package sizes for clients for downloading cost estimation
        self.gv.simulator.set_variable(target_id, '__download_package_size',size_of_package(package))
        res = communicate_with(self, target_id, package)
        # Set uploading package sizes for clients for uploading cost estimation
        self.gv.simulator.set_variable(target_id, '__upload_package_size', size_of_package(res))
        # update latency of the target client according to the communication cost and computation cost
        self.gv.simulator.update_client_responsiveness([target_id])
        # Record the size of the package that may influence the value of the latency
        # Update the real-time latency of the client response
        # Get the updated latency
        latency = self.gv.simulator.get_variable(target_id, 'latency')[0]
        self.clients[target_id]._latency = latency
        res['__cid'] = target_id
        # Compute the arrival time
        res['__t'] = self.gv.clock.current_time + latency
        return res
    return delayed_communicate_with

# local_movielens_recommendation training phase
def with_completeness(train):
    r"""
    The decorator to simulate the scene where the clients may upload incomplete model updates

    Example:
    ```python
        >>> import flgo.algorithm.fedbase
        >>> import flgo.simulator.base as ss
        >>> class Client(flgo.algorithm.fedbase.BasicClient):
        ...     @ss.with_completeness
        ...     def train(self,...):
        ...         ...
    ```
    """
    @functools.wraps(train)
    def train_with_incomplete_update(self, model, *args, **kwargs):
        old_num_steps = self.num_steps
        self.num_steps = self._working_amount
        res = train(self, model, *args, **kwargs)
        self.num_steps = old_num_steps
        return res
    return train_with_incomplete_update

def with_clock(communicate):
    r"""
    The decorator to simulate the scene where there is a virtual global clock

    Example:
    ```python
        >>> import flgo.algorithm.fedbase
        >>> import flgo.simulator.base as ss
        >>> class Server(flgo.algorithm.fedbase.BasicServer):
        ...     @ss.with_clock
        ...     def communicate(self,...):
        ...         ...
    ```
    """
    def communicate_with_clock(self, selected_clients, mtype=0, asynchronous=False):
        self.gv.simulator.update_client_completeness(selected_clients)
        res = communicate(self, selected_clients, mtype, asynchronous)
        # If all the selected clients are unavailable, directly return the result without waiting.
        # Else if all the available clients have dropped out and not using asynchronous communication,  waiting for `tolerance_for_latency` time units.
        tolerance_for_latency = self.get_tolerance_for_latency()
        if not asynchronous and len(selected_clients)==0:
            if hasattr(self, '_dropped_selected_clients') and len(self._dropped_selected_clients)>0:
                self.gv.clock.step(tolerance_for_latency)
            return res
        # Convert the unpacked packages to a list of packages of each client.
        pkgs = [{key: vi[id] for key, vi in res.items()} for id in range(len(list(res.values())[0]))] if len(selected_clients)>0 else []
        # Put the packages from selected clients into clock only if when there are effective selected clients
        if len(selected_clients)>0:
            # Set selected clients' states as `working`
            self.gv.simulator.set_client_state(selected_clients, 'working')
            for pi in pkgs: self.gv.clock.put(pi, pi['__t'])
        # Receiving packages in asynchronous\synchronous way
        # Wait for client packages. If communicating in asynchronous way, the waiting time is 0.
        if asynchronous:
            # Return the currently received packages to the server
            eff_pkgs = self.gv.clock.get_until(self.gv.clock.current_time)
            eff_cids = [pkg_i['__cid'] for pkg_i in eff_pkgs]
        else:
            # Wait all the selected clients for no more than `tolerance_for_latency` time units.
            # Check if anyone had dropped out or will be overdue
            max_latency = max(self.gv.simulator.get_variable(selected_clients, 'latency'))
            any_drop, any_overdue = (len(self._dropped_selected_clients) > 0), (max_latency >  tolerance_for_latency)
            # Compute delta of time for the communication.
            delta_t = tolerance_for_latency if any_drop or any_overdue else max_latency
            # Receive packages within due
            eff_pkgs = self.gv.clock.get_until(self.gv.clock.current_time + delta_t)
            self.gv.clock.step(delta_t)
            # Drop the packages of overdue clients and reset their states to `idle`
            eff_cids = [pkg_i['__cid'] for pkg_i in eff_pkgs]
            self._overdue_clients = list(set([cid for cid in selected_clients if cid not in eff_cids]))
            # no additional wait for the synchronous selected clients and preserve the later packages from asynchronous clients
            if len(self._overdue_clients) > 0:
                self.gv.clock.conditionally_clear(lambda x: x['__cid'] in self._overdue_clients)
                self.gv.simulator.set_client_state(self._overdue_clients, 'idle')
            # Resort effective packages
            pkg_map = {pkg_i['__cid']: pkg_i for pkg_i in eff_pkgs}
            eff_pkgs = [pkg_map[cid] for cid in selected_clients if cid in eff_cids]
        self.gv.simulator.set_client_state(eff_cids, 'offline')
        self.received_clients = [pkg_i['__cid'] for pkg_i in eff_pkgs]
        return self.unpack(eff_pkgs)
    return communicate_with_clock





from torch import nn
from flgo.utils.fmodule import FModule
import torch
from torchtext.data.utils import get_tokenizer
from torchtext.vocab import build_vocab_from_iterator
from torchtext.datasets import AG_NEWS

def yield_tokens(data_iter):
    for _, text in data_iter:
        yield tokenizer(text)

tokenizer = get_tokenizer('basic_english')
train_iter = AG_NEWS(split='train')

vocab = build_vocab_from_iterator(yield_tokens(train_iter), specials=["<unk>"])
vocab.set_default_index(vocab["<unk>"])

text_pipeline = lambda x: vocab(tokenizer(x))
label_pipeline = lambda x: int(x) - 1

def collate_fn(batch):
    label_list, text_list, offsets = [], [], [0]
    for (_label, _text) in batch:
         label_list.append(label_pipeline(_label))
         processed_text = torch.tensor(text_pipeline(_text), dtype=torch.int64)
         text_list.append(processed_text)
         offsets.append(processed_text.size(0))
    label_list = torch.tensor(label_list, dtype=torch.int64)
    offsets = torch.tensor(offsets[:-1]).cumsum(dim=0)
    text_list = torch.cat(text_list)
    return label_list, text_list, offsets

class Model(FModule):
    def __init__(self, vocab_size=len(vocab), embed_dim=64, num_class=4):
        super(Model, self).__init__()
        self.embedding = nn.EmbeddingBag(vocab_size, embed_dim, sparse=False)
        self.fc = nn.Linear(embed_dim, num_class)
        self.init_weights()

    def init_weights(self):
        initrange = 0.5
        self.embedding.weight.data.uniform_(-initrange, initrange)
        self.fc.weight.data.uniform_(-initrange, initrange)
        self.fc.bias.data.zero_()

    def forward(self, text, offsets):
        embedded = self.embedding(text, offsets)
        return self.fc(embedded)

def init_dataset(object):
    object.calculator.set_collect_fn(collect_fn=collate_fn)

def init_local_module(object):
    return

def init_global_module(object):
    if 'Server' in object.get_classname():
        model = Model().to(object.device)
        object.set_model(model)


from flgo.benchmark.shakespeare_classification.model import stackedlstm
import flgo.benchmark.toolkits.partition

default_partitioner = flgo.benchmark.toolkits.partition.IIDPartitioner
default_partition_para = {'num_clients':100}
default_model = stackedlstm


from .model import bag_linear
import flgo.benchmark.toolkits.partition

default_partitioner = flgo.benchmark.toolkits.partition.IIDPartitioner
default_partition_para = {'num_clients':10}
default_model = bag_linear